# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pandas as pd
import matplotlib.pyplot as plt

from highcharts import Highchart
from typing import Dict, Optional, Sequence, Tuple
from ..utils.logging import log
"""
A module for data visualization
"""

__all__ = [
    "getHeatColor", "convertHexToRGB", "convertRGBToHex", "plotScatter", "plotGrid", "plotAUC"
]


def getHeatColor(intensity: float,
                 topColor: str,
                 bottomColor: str = "#ffffff") -> str:
    """
    Calculation of heat map color values based on intensity values

    Parameters
    ----------
    intensity: float value
        color intensity values between 0 and 1
    topColor: hexadecimal color string
       Color value when intensity value is 1
    bottomColor: hexadecimal color string
     Color value when intensity value is 0, default is white

    Returns
    ----------
        The hexadecimal color string corresponding to the intensity

    Examples
    ----------
    >>> getHeatColor(0.5, "#ff0000")
    '#ff7f7f'
    """
    assert intensity >= 0 and intensity <= 1 + 1e-8, f"Intensity should range from 0 to 1, But got {intensity}"
    bRGBColor = convertHexToRGB(bottomColor)
    tRGBColor = convertHexToRGB(topColor)
    color = ((1 - intensity) * bRGBColor[i] + intensity * tRGBColor[i]
             for i in range(3))
    return convertRGBToHex(tuple(color))


def convertHexToRGB(hex_color: str) -> Tuple[int]:
    """
    Convert hexadecimal color strings to RGB tri-color integer tuples

    Parameters
    ----------
    hex_color: hexadecimal color string, such as '#ff0000'

    Returns
    ----------
    RGB tri-color integer tuples

    Examples
    ----------
    >>> hexToRGB('#ff0000')
    (255, 0, 0)
    """
    if not isinstance(hex_color, str):
        raise ValueError("hex_color should be a string")
    if hex_color[0] != "#":
        raise ValueError("hex_color should start with '#'")
    if len(hex_color) != 7:
        raise ValueError("hex_color should be 7 characters long")

    try:
        d = int(hex_color.replace("#", ""), base=16)
        b, d = d % 256, d // 256
        g, d = d % 256, d // 256
        r, d = d % 256, d // 256
        return r, g, b
    except Exception:
        raise ValueError("Ilegal hex color format")


def convertRGBToHex(color: Tuple[int]) -> str:
    """
    Convert RGB tricolor integer tuple to hexadecimal color string

    Parameters
    ----------
    color: RGB tricolor integer tuple
        such as (255, 0, 0)

    Returns
    ----------
    hexadecimal color string, such as '#ff0000'

    Examples
    ----------
    >>> rgbToHex((255, 0, 0))
    '#ff0000'
    """
    assert color is not None and len(color) == 3, "Ilegal rgb color format"
    try:
        r = int(color[0] % 256)
        g = int(color[1] % 256)
        b = int(color[2] % 256)
        return "#%02x%02x%02x" % (r, g, b)
    except Exception:
        raise ValueError("Ilegal rgb color format")


# see https://api.highcharts.com/highcharts
def plotScatter(data2d: pd.DataFrame,
                colormap: Dict[str, str],
                title: str,
                subtitle: str,
                filename: str = None,
                scatter_symbol: str = "circle",
                width: int = 1000,
                height: int = 850,
                radius: int = 3,
                x_min: Optional[int] = None,
                y_min: Optional[int] = None,
                x_max: Optional[int] = None,
                y_max: Optional[int] = None,
                x_title: Optional[str] = None,
                y_title: Optional[str] = None):
    """
    Draw the scatter image of socube

    Parameters
    ----------
    data2d: pandas.DataFrame
        The data to be plotted, with columns of x, y, label and subtype,
        if subtype is float, it regarded as intensity value, and the color will be
        calculated based on the intensity value.
    colormap: Dict[str, str]
        The color map for the subtype, with key as subtype name and value
        as hexadecimal color string. If the subtype is float, colormap's key
        should be '0' and '1', '0' for low intensity color and '1' for high intensity color.
    title: str
        The title of the plot
    subtitle: str
        The subtitle of the plot
    filename: str
        The filename of the plot, if None, the plot will not be saved. format is html and
        the filename extension will automatically be added and you should not add it.
    width: int
        The width of the plot, unit is pixel
    height: int
        The height of the plot, unit is pixel
    radius: int
        The radius of the scatter point, unit is pixel

    Returns
    ----------
    The plot object

    Examples
    ----------
    >>> data2d = pd.DataFrame({'x': [1, 2, 3], 'y': [4, 5, 6], 'label': ['a', 'b', 'c'], 'subtype': [0.5, 0.7, 0.9]})
    >>> colormap = {'0': '#ff0000', '1': '#00ff00'}
    >>> plotScatter(data2d, colormap, 'title', 'subtitle', 'test.html')
    """
    data2d = data2d.copy()
    # Create drawing objects and configure global parameters
    H = Highchart(width=width, height=height)

    # Set jquery CDN to load
    H.JSsource[0] = "http://code.jquery.com/jquery-1.9.1.min.js"
    H.set_options('chart', {'type': 'scatter', 'zoomType': 'xy'})
    H.set_options('title', {'text': title})
    H.set_options('subtitle', {'text': subtitle})
    H.set_options(
        'xAxis',
        {
            'title': {
                'enabled': True,
                'text': x_title,
                'style': {
                    'fontSize': 20
                }
            },
            'labels': {
                'style': {
                    'fontSize': 20
                }
            },
            'showLastLabel': True,
            'min': x_min,
            'max': x_max
        })
    H.set_options(
        'yAxis',
        {
            'title': {
                'text': y_title,
                'style': {
                    'fontSize': 20
                }
            },
            'labels': {
                'style': {
                    'fontSize': 20
                }
            },
            'min': y_min,
            'max': y_max
        })
    H.set_options(
        'legend', {
            'align': 'right',
            'layout': 'vertical',
            'margin': 1,
            'verticalAlign': 'top',
            'y': 40,
            'symbolHeight': 12,
            'floating': False,
        })
    H.set_options(
        'plotOptions', {
            'scatter': {
                'marker': {
                    'radius': radius,
                    'states': {
                        'hover': {
                            'enabled': True,
                            'lineColor': 'rgb(100,100,100)'
                        }
                    },
                    "symbol": scatter_symbol
                },
                'states': {
                    'hover': {
                        'marker': {
                            'enabled': False
                        }
                    }
                },
                'tooltip': {
                    'headerFormat': '<b>{series.name}</b><br>',
                    'pointFormat': 'Label: {point.label}'
                }
            },
            'series': {
                'turboThreshold': 50000
            }
        })

    if data2d.subtype.dtype.name == "object":
        # Add series by subtype
        for subtype, color in colormap.items():
            dfi = data2d[data2d.subtype.apply(
                lambda subtypes: subtype in subtypes.split("||"))]
            if len(dfi) == 0:
                continue
            data = dfi.to_dict('records')
            H.add_data_set(data, 'scatter', subtype, color=color)
    elif "float" in data2d.subtype.dtype.name:
        data2d["color"] = data2d.subtype.apply(lambda intensity: getHeatColor(
            intensity, colormap["1"], colormap["0"]))
        H.add_data_set(data2d.to_dict('records'), 'scatter')
    else:
        raise NotImplementedError("Unsupport 'subtype' column's dtype")

    if filename:
        H.save_file(filename)
        log(__name__, f"Save to {filename}.html")

    return H


def plotGrid(data2d: pd.DataFrame,
             colormap: Dict[str, str],
             shape: Tuple[int],
             title: str,
             subtitle: str,
             filename: str,
             width: int = 1000,
             height: int = 850) -> Highchart:
    """
    Draw socube's Grid image

    Parameters
    ----------
    data2d: pandas.DataFrame
        The data to be plotted, with columns of x, y, label and subtype,
        if subtype is float, it regarded as intensity value, and the color will be
        calculated based on the intensity value.
    colormap: Dict[str, str]
        The color map for the subtype, with key as subtype name and value
        as hexadecimal color string. If the subtype is float, colormap's key
        should be '0' and '1', '0' for low intensity color and '1' for high intensity color.
    shape: Tuple[int]
        The shape of the grid, (row, col)
    title: str
        The title of the plot
    subtitle: str
        The subtitle of the plot
    filename: str
        The filename of the plot, if None, the plot will not be saved. format is html and
        the filename extension will automatically be added and you should not add it.
    width: int
        The width of the plot, unit is pixel
    height: int
        The height of the plot, unit is pixel

    Returns
    ----------
    The plot object

    Examples
    ----------
    >>> data2d = pd.DataFrame({'x': [1, 2, 3], 'y': [4, 5, 6], 'label': ['a', 'b', 'c'], 'subtype': [0.5, 0.7, 0.9]})
    >>> colormap = {'0': '#ff0000', '1': '#00ff00'}
    >>> plotGrid(data2d, colormap, (6, 3), 'title', 'subtitle', 'test.html')
    """
    data2d = data2d.copy()
    H = Highchart(width=width, height=height)
    H.JSsource[0] = "http://code.jquery.com/jquery-1.9.1.min.js"
    H.set_options('chart', {'type': 'heatmap', 'zoomType': 'xy'})
    H.set_options('title', {'text': title})
    H.set_options('subtitle', {'text': subtitle})
    H.set_options(
        'xAxis', {
            'title': None,
            'min': 0,
            'max': shape[1],
            'startOnTick': False,
            'endOnTick': False,
            'allowDecimals': False,
            'labels': {
                'style': {
                    'fontSize': 20
                }
            }
        })
    H.set_options(
        'yAxis', {
            'title': {
                'text': ' ',
                'style': {
                    'fontSize': 20
                }
            },
            'startOnTick': False,
            'endOnTick': False,
            'gridLineWidth': 0,
            'reversed': True,
            'min': 0,
            'max': shape[0],
            'allowDecimals': False,
            'labels': {
                'style': {
                    'fontSize': 20
                }
            }
        })
    H.set_options(
        'legend', {
            'align': 'right',
            'layout': 'vertical',
            'margin': 1,
            'verticalAlign': 'top',
            'y': 60,
            'symbolHeight': 12,
            'floating': False,
        })
    H.set_options(
        'tooltip', {
            'headerFormat': '<b>{series.name}</b><br>',
            'pointFormat': '{point.label} {point.x} {point.y}'
        })

    # plotOptions.series.turboThreshold is the threshold of serie point.
    # For detail , See https://api.highcharts.com/highcharts/plotOptions.series.turboThreshold
    H.set_options('plotOptions', {'series': {'turboThreshold': 50000}})
    if data2d.subtype.dtype.name == "object":
        for subtype, color in colormap.items():
            dfi = data2d[data2d.subtype.apply(
                lambda subtypes: subtype in subtypes.split("||"))]
            if len(dfi) == 0:
                continue
            data = dfi.to_dict('records')
            H.add_data_set(data, 'heatmap', subtype, color=color)
    elif "float" in data2d.subtype.dtype.name:
        data2d["color"] = data2d.subtype.apply(lambda intensity: getHeatColor(
            intensity, colormap["1"], colormap["0"]))
        H.add_data_set(data2d.to_dict('records'), 'heatmap')
    else:
        raise NotImplementedError("Unsupport 'subtype' column's dtype")

    if filename:
        H.save_file(filename)
        log(__name__, f"Save to {filename}.html")
    return H


def plotAUC(data: Dict[str, Tuple[Sequence]],
            title: str,
            xlabel: str,
            ylabel: str,
            file: Optional[str] = None,
            slash: int = 0):
    """
    Plot a AUC curve

    Parameters
    ------------------
    data: Dict[str, Tuple[Sequence]]
        The data to be plotted, with key as the name of the curve and value
        as the (x, y) data.
    title: str
        The title of the plot
    xlabel: str
        The xlabel of the plot
    ylabel: str
        The ylabel of the plot
    file: str
        The filename of the plot, if None, the plot will not be saved
    slash: int
        if `slash` is positve, will plot a forward slash,
        if `slash` is negative, will plot a back slash.

    Examples
    ----------
    >>> data = {'AUC': ([0.1, 0.2, 0.3], [0.1, 0.2, 0.3]), 'AUC2': ([0.1, 0.2, 0.3], [0.1, 0.2, 0.3])}
    >>> plotAUC(data, 'AUC', 'xlabel', 'ylabel', 'auc.png')
    """
    # reuse figure object to avoid open too many figures and cosume too much memory
    plt.figure(1024)
    for legend, vals in data.items():
        plt.plot(vals[0], vals[1], lw=2, label=legend)

    if slash > 0:
        plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
    elif slash < 0:
        plt.plot([0, 1], [1, 0], color='navy', lw=2, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.legend(loc="lower right")
    if file is None:
        return
    plt.savefig(file)

    # clean current activate figure for reusing.
    plt.clf()
    plt.close(1024)



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from typing import (Optional, Tuple, TypeVar, Union)
import pandas as pd
import numpy as np
import torch
import abc

from os.path import join

from sklearn.decomposition import PCA
from .data import (std, minmax, scatterToGrid, tsne2D, umap2D)
from .utils.logging import log, getJobId
from .utils.io import (writeHdf, mkDirs, writeCsv, writeNpy, checkExist)
from .utils.concurrence import ParallelManager
from tqdm.auto import tqdm

__all__ = ["SoCube"]
SoCubeType = TypeVar("SoCubeType", bound="SoCube")


class SoCube(metaclass=abc.ABCMeta):
    """
    Abstract class of SoCube.

    Parameters
    ----------
    data_path : str
        Path of original scRNA-seq data file
    home_path : str, optional
        Path of output home directory. If not specified, it will be set to
        the same directory as data_path.
    metric: str, optional
        Metric to use for distance calculation. supported:
        'correlation', 'cosine', 'euclidean', 'manhattan', etc.
        default is 'correlation'.
    dtype: str, optional
        Data type of output data. isupported: 'float32', 'float64', etc.
        default is 'float32'.
    verbose: int, optional
        Verbose level. default is 1.
    """

    def __init__(self,
                 data_path: str,
                 home_path: Optional[str] = None,
                 metric: str = 'correlation',
                 dtype: str = "float32",
                 verbose: int = 1,
                 **kwargs) -> None:
        checkExist(data_path, "dir")
        if home_path is None:
            home_path = data_path

        embedding_path = join(home_path, "embedding")
        mkDirs(embedding_path)
        self._config = {
            "data_path": data_path,
            "embedding_path": embedding_path,
            "metric": metric,
            "dtype": dtype,
            "home_path": home_path,
            "verbose": verbose,
            "socube": self.__class__.__name__
        }
        self._config.update(kwargs)

        self._manifold_methods = {"umap": umap2D, "tsne": tsne2D}
        self._grid = None
        self._scatter = None
        self._shape = None

    """Abastract map class"""

    @abc.abstractmethod
    def fit(self, **kwargs):
        """all subclass must implement it"""
        for k, v in kwargs.items():
            self._config[k] = v

    @staticmethod
    def instance(map_type: str,
                 data_path: str,
                 home_path: Optional[str] = None,
                 metric: str = 'correlation',
                 dtype: str = "float32",
                 verbose: int = 1,
                 **kwargs) -> SoCubeType:
        """
        SoCube is designed using the factory pattern, and this method
        is used to create SoCube instance objects

        Parameters
        ----------
        map_type : str
            Type of map to use. supported: 'pca'
        data_path : str
            Path of original scRNA-seq data file
        home_path : str, optional
            Path of output home directory. If not specified, it will be set to
            the same directory as data_path.
        metric: str, optional
            Metric to use for distance calculation. supported:
            'correlation', 'cosine', 'euclidean', 'manhattan', etc.
            default is 'correlation'.
        dtype: str, optional
            Data type of output data. isupported: 'float32', 'float64', etc.
            default is 'float32'.
        verbose: int, optional
            Verbose level. default is 1.

        Returns
        -------
        SoCube instance object
        """
        avail = SoCube.available()
        if map_type in avail:
            return avail[map_type](data_path=data_path,
                                   home_path=home_path,
                                   metric=metric,
                                   dtype=dtype,
                                   verbose=verbose,
                                   **kwargs)
        else:
            raise TypeError(f"Only support: {avail.keys()}")

    @staticmethod
    def available() -> dict:
        """
        Return available map types
        """
        return {"pca": SoPCACube}

    def _log(self, message: str, verbose_level: int = 1, **kwargs):
        """
        The logging methods used inside the SoCube class are
        wrappers for the socube.utils.logging module

        Parameters
        ----------
        message : str
            Message to be logged
        verbose_level : int, optional
            Verbose level. default is 1. The higher the verbose value
            the higher the priority of the log
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the logging module
        """
        log(self.__class__.__name__,
            message,
            quiet=self._config["verbose"] > verbose_level,
            **kwargs)

    def _std(self, data: pd.DataFrame, file: str = None) -> pd.DataFrame:
        """
        Standardize the data. A wrapper for socube.data.std

        Parameters
        ----------
        data : pd.DataFrame
            Data to be standardized
        file : str, optional
            Path to save standardized data. If not specified, the standardized
            data will not be saved.

        Returns
        -------
        pd.DataFrame
            Standardized data
        """
        self._log("Standard data by z-score")
        data = std(data, dtype=self._config["dtype"])
        if file is not None:
            writeHdf(data,
                     file,
                     callback=lambda: self._log("Standard data saved"))
        return data

    def _normalize(self, data: pd.DataFrame, file: str = None) -> pd.DataFrame:
        """
        Normalize the data. A wrapper for socube.data.minmax

        Parameters
        ----------
        data : pd.DataFrame
            Data to be normalized
        file : str, optional
            Path to save normalized data. If not specified, the normalized
            data will not be saved.

        Returns
        -------
        pd.DataFrame
            Normalized data
        """
        self._log("Normalize data by min-max scale")
        data = minmax(data, dtype=self._config["dtype"])
        if file is not None:
            writeHdf(data,
                     file,
                     callback=lambda: self._log("Normalized data saved"))
        return data

    def _alignment(self, data: pd.DataFrame, file: str = None) -> pd.DataFrame:
        """
        Align the data. A wrapper for socube.data.scatterToGrid

        Parameters
        ----------
        data : pd.DataFrame
            Data to be aligned, it must be a dataframe with columns 'x' and 'y'
            which are the x and y coordinates of the data in the scatter plot
        file : str, optional
            Path to save aligned data. If not specified, the aligned
            data will not be saved.

        Returns
        -------
        pd.DataFrame
            Aligned data, which is a dataframe with two columns 'x' and 'y',
            representing the x and y coordinates of the data in the grid
        """
        assert data.shape[1] == 2, f"Expect 2 col but got {data.shape[1]}"
        self._log("JV alignment to grid")
        grid_tensor = scatterToGrid(torch.from_numpy(data.values),
                                    device_name=self._config["device"])
        grid_data = pd.DataFrame(grid_tensor.numpy(),
                                 index=data.index,
                                 columns=["x", "y"])

        # grid shape row * col
        self.shape = (int(grid_data.y.max()) + 1, int(grid_data.x.max()) + 1)
        self._log(f"Grid shape is {self.shape}")
        if file is not None:
            writeCsv(grid_data,
                     file,
                     callback=lambda: self._log("Grid data saved"))
        return grid_data

    @property
    def grid(self) -> pd.DataFrame:
        """Return the grid data"""
        assert isinstance(self._grid, pd.DataFrame), "Uninitialized grid data"
        return self._grid.copy()

    @grid.setter
    def grid(self, data: pd.DataFrame):
        """Set the grid data"""
        assert isinstance(data, pd.DataFrame) and len(
            data.shape) == 2, "Invalid grid data"
        self._grid = data.copy()
        self._grid.columns = ["x", "y"]
        self.shape = (int(data.y.max()) + 1, int(data.x.max()) + 1)

    @property
    def scatter(self) -> pd.DataFrame:
        """Return the scatter data"""
        assert isinstance(self._scatter,
                          pd.DataFrame), "Uninitialized scatter data"
        return self._scatter.copy()

    @scatter.setter
    def scatter(self, data: pd.DataFrame):
        """Set the scatter data"""
        assert isinstance(data, pd.DataFrame) and len(
            data.shape) == 2, "Invalid scatter data"
        self._scatter = data.copy()

    @property
    def shape(self) -> tuple:
        """
        Return the socube 2d grid's shape
        """
        assert self._shape, "Uninitialized shape property"
        return self._shape

    @shape.setter
    def shape(self, new_shape: Tuple[int]):
        """
        Set the socube 2d grid's shape.
        """
        new_shape = tuple(filter(lambda obj: isinstance(obj, int), new_shape))
        assert len(new_shape) == 2, "Invalid grid shape change"
        self._shape = (new_shape[0], new_shape[1])

    def _manifold(self,
                  data: pd.DataFrame,
                  methods: str = 'umap',
                  file: str = None) -> pd.DataFrame:
        """
        Dimensionality reduction to two-dimensional space using
        manifold learning algorithm

        Parameters
        ----------
        data : pd.DataFrame
            Data to be reducted, rows represent genes and cols
            represent cells
        file : str, optional
            Path to save reducted scatter data. If not specified,
            the data will not be saved.

        Returns
        -------
        pd.DataFrame
            Aligned data, which is a dataframe with two columns 'x' and 'y',
            representing the x and y coordinates of the data in the scatter
        """
        assert methods in self._manifold_methods, "Unsupport manifold methods"
        self._log(f"Manifold learning by {methods}")
        data = self._manifold_methods[methods](data, self._config["metric"])

        if file is not None:
            writeCsv(data,
                     file,
                     callback=lambda: self._log("Manifold 2d result saved"))
        return data

    @abc.abstractmethod
    def transform(self, sample: pd.Series, **kwargs) -> np.ndarray:
        """
        Transform a droplet sample vector to
        its embedding form. for batch transform,
        see `batchTransform` api.

        Parameters
        ----------
        sample: pd.Series
            a droplet sample vector

        Returns
        ----------
        embedding form
        """
        raise NotImplementedError

    def batchTransform(self,
                       samples: pd.DataFrame,
                       data_dir: Optional[str] = None,
                       use_index: bool = False,
                       **kwargs) -> np.ndarray:
        r"""
        The sample data batch is subjected to feature transformation
        by socube to obtain the feature 3d matrix in the form of NCHW

        Parameters
        ----------------
        samples: pd.DataFrame
            The sample data to be transformed, row as samples, columns as features
        data_dir: str
            Path to save samples.  If given, it will save samples
        use_index: bool
            If True, it will save samples with using numerical index as filename. such as `0.npy`
            otherwise, sample name used instead.

        Returns
        -----------
            The generated embedding feature matrix and gene matching
        """
        gene_match = [0, 0]
        if data_dir is not None:
            mkDirs(data_dir)

        with ParallelManager(verbose=True):

            def _transform(sample: pd.Series):
                res, match = self.transform(sample, **kwargs)
                gene_match[0] = match[0]
                gene_match[1] = match[1]
                if data_dir is not None:
                    if use_index:
                        name = join(
                            data_dir,
                            f"{samples.index.get_loc(sample.name)}.npy")
                    else:
                        name = join(data_dir, f"{sample.name}.npy")
                    writeNpy(res, name)
                return res

            tqdm.pandas(desc="Transform")
            self._log(f"Batch transform {samples.shape[0]} samples")
            res = np.stack(samples.progress_apply(_transform, axis=1))
        self._log("Gene matched:{:d}, not matched:{:d}".format(*gene_match))
        return res, gene_match

    @abc.abstractmethod
    def load(self, cube_path: str) -> bool:
        """
        Load pretrained Socube
        """
        raise NotImplementedError


class SoPCACube(SoCube):
    """
    SoCube implemented with PCA latent gene feature
    """

    def __init__(self,
                 data_path: str,
                 home_path: Optional[str] = None,
                 metric: str = 'correlation',
                 dtype: str = "float32",
                 verbose: int = 1) -> None:

        super(SoPCACube, self).__init__(data_path=data_path,
                                        metric=metric,
                                        dtype=dtype,
                                        home_path=home_path,
                                        verbose=verbose)
        self._latent = None

    def fit(self,
            data: Union[str, pd.DataFrame],
            latent_dim: int,
            with_log: bool = True,
            with_std: bool = True,
            with_norm: bool = True,
            device: str = "cpu",
            manifold: str = 'umap',
            job_id: Optional[str] = None,
            seed: Optional[int] = None) -> str:
        """
        Fitting the most suitable cube representation to
        the currently specified data

        Parameters
        -------------
        data: str or pd.DataFrame
            If a str passed, it will be regarded as the input data's filename,
            otherwise a dataframe passed as the input data.
        latent_dim: int
            gene latent feature's target dimension. it is the third
            dimension of socube.
        with_log: bool
            Whether process data with logarithmic transform.
        with_std: bool
            Whether standardized data
        with_norm: bool
            Whether normalized data
        device: str
            Device name. cpu or cuda
        manifold: str
            Manifold algorithm, only 'tsne' and 'umap' supported
        job_id: str
            Job id, if is not specified, will be automatically generated.
        seed: int
            Random seed

        Returns
        -------
        Job id returned
        """
        if job_id is None:
            job_id = getJobId()
        super(SoPCACube, self).fit(jobid=job_id,
                                   device=device,
                                   log=with_log,
                                   std=with_std,
                                   norm=with_norm,
                                   latent_dim=latent_dim,
                                   manifold=manifold,
                                   seed=seed)
        if isinstance(data, str):
            self._config["data"] = data

        self._config["embedding_path"] = join(self._config["embedding_path"],
                                              job_id)
        config_path = join(self._config["embedding_path"], "config.json")
        json.dump(self._config, open(config_path, "w"))
        self._log("Current socube job config will be saved to {}".format(
            config_path))

        mkDirs(self._config["embedding_path"])

        # read data
        if isinstance(data, str):
            infile = join(self._config["data_path"], data)
            checkExist(infile)
            self._log(f"Load data from {infile}")
            data = pd.read_hdf(infile)
            self._log("Data loaded")
        elif not isinstance(data, pd.DataFrame):
            raise TypeError(f"Invalid data type '{type(data)}'")

        if with_log:
            data = np.log(data + 1)

        if with_std:
            data = self._std(data,
                             file=join(self._config["embedding_path"],
                                       "std.h5"))

        if with_norm:
            data = self._normalize(data,
                                   file=join(self._config["embedding_path"],
                                             "norm.h5"))

        data = data.sample(frac=1, random_state=seed)
        self.latent = self._pca(data,
                                latent_dim,
                                seed=seed,
                                file=join(self._config["embedding_path"],
                                          "latent.h5"))

        self.scatter = self._manifold(data,
                                      manifold,
                                      file=join(self._config["embedding_path"],
                                                "scatter.csv"))

        self.grid = self._alignment(self.scatter,
                                    file=join(self._config["embedding_path"],
                                              "grid.csv"))
        return job_id

    def _pca(self,
             data: pd.DataFrame,
             latent_dim: int,
             seed: Optional[int] = None,
             file: Optional[str] = None) -> pd.DataFrame:
        """
        PCA reduction used for getting gene latent feature

        Parameters
        ----------
        data: pd.DataFrame
            Input data
        latent_dim: int
            Target latent feature dimension
        seed: int
            Random seed
        file: str
            Npy file name to save the latent feature

        Returns
        -------
        pd.DataFrame
            Latent feature
        """
        pca = PCA(n_components=latent_dim, random_state=seed)
        transform = pca.fit_transform(data.values)
        transform = pd.DataFrame(transform, index=data.index)
        if file is not None:
            self._log("Write pca result to file")
            writeHdf(transform,
                     file,
                     callback=lambda: self._log("Gene pca data saved"))
        return transform

    @property
    def latent(self) -> pd.DataFrame:
        """Return latent feature"""
        assert isinstance(self._latent,
                          pd.DataFrame), "Uninitialize gene latent"
        return self._latent.copy()

    @latent.setter
    def latent(self, data: pd.DataFrame):
        """Set latent feature"""
        assert isinstance(data, pd.DataFrame) and len(
            data.shape) == 2, "Invalid latent"
        self._latent = data.copy()

    def transform(self, sample: pd.Series) -> np.ndarray:
        """Implementation of load method"""
        df = self.grid.join(sample)
        counts: dict = df.iloc[:, 2].isna().value_counts().to_dict()
        geneMatch = (counts.setdefault(False, 0), counts.setdefault(True, 0))
        df = df.fillna(0)
        df.columns = ['col', 'row', 'val']
        latent2d: np.ndarray = self.latent.loc[df.index].values
        # negative means this dim infered by others
        # weighted calculation
        latent2d *= df.val.values.reshape((-1, 1))
        # imputate and convert to 3D
        shape = self.shape
        latent3d = np.concatenate([
            latent2d,
            np.zeros(
                (shape[0] * shape[1] - latent2d.shape[0], latent2d.shape[1]))
        ]).reshape((*shape, -1))
        # sorted gene by idx, imputation pixel always at the end of matrix and no influence to sort
        latent3d[df.row.values, df.col.values] = latent2d
        # Convert HWC to CHW
        latent3d = np.transpose(latent3d, [2, 0, 1])
        return latent3d, geneMatch

    def load(self, cube_path: str) -> bool:
        """Implementation of load method"""
        try:
            self.grid = pd.read_csv(join(cube_path, "grid.csv"), index_col=0)
            self.scatter = pd.read_csv(join(cube_path, "scatter.csv"),
                                       index_col=0)
            self.latent = pd.read_hdf(join(cube_path, "latent.h5"))
            self._config = json.load(open(join(cube_path, "config.json")))
            self._log(f"Load socube job from {cube_path}")
            return True
        except Exception as e:
            self._log(f"Load socube failed: {e}")
            return False



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import time
import random
"""This module provides some logging tools"""
__all__ = [
    "log", "getJobId", "DEBUG_LEVEL", "INFO_LEVEL", "WARN_LEVEL", "ERROR_LEVEL"
]

DEBUG_LEVEL = "debug"
INFO_LEVEL = "info"
WARN_LEVEL = "warn"
ERROR_LEVEL = "error"

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


def log(name: str,
        msg: str,
        level: str = "info",
        quiet: bool = False,
        *args,
        **kwargs) -> None:
    """
    Log a message.

    Parameters
    ----------
    name : str
        The name of the logger.
    msg : str
        The message to be logged.
    level : str
        The level of the message.
    quiet : bool
        If True, the message will not be logged.
    *args :
        The rest of the arguments will be passed to the logger.
    **kwargs :
        The rest of the keyword arguments will be passed to the logger.

    Examples
    ----------
    >>> log("mylogger", "hello world", level="debug")
    >>> log("mylogger", "hello world", level="info")
    >>> log("mylogger", "hello world", level="warn")
    >>> log("mylogger", "hello world", level="error")
    """
    if quiet:
        return
    logger = logging.getLogger(name)
    log_dict = {
        "debug": logger.debug,
        "info": logger.info,
        "warn": logger.warning,
        "error": logger.error
    }
    assert level in log_dict, f"Invalid level {level}"
    log_dict[level](msg, *args, **kwargs)


def getJobId() -> str:
    """
    Generate job id randomly. It's useful to seperate different jobs.

    Returns
    ----------
        a string id

    Examples
    ----------
    >>> getJobId()
    '20211102-141041-237'

    >>> getJobId()
    '20211102-132411-806'
    """
    s = time.strftime(r"%Y%m%d-%H%M%S", time.localtime())
    return "%s-%d" % (s, random.randint(100, 999))



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import abc
import torch

from torch.nn import Module
from torch.nn.parameter import Parameter
from typing import Any, Iterator, Tuple, Type

__all__ = ["ModuleMetaBase", "ModuleBase", "NetBase"]


class ModuleMetaBase(abc.ABCMeta):
    """
    Metaclass for all neural network model implemented in this package.

    Parameters
    ----------
    cls : type
        The class being created, which is a instance of ModuleMetaBase and
        a subclass of torch.nn.Module.
    in_channels : int
        The number of input channels. Any type created by this metaclass
        must contain this constructor parameter.
    out_channels : int
        The number of output channels. Any type created by this metaclass
        must contain this constructor parameter.
    *args : Any
        The arguments for the class constructor
    **kwargs : Any
        The keyword arguments for the class constructor

    Returns
    -------
    A new instance of the class being created.
    """
    def __call__(cls: Type[Module], in_channels: int, out_channels: int,
                 *args: Any, **kwargs: Any) -> Module:
        if not issubclass(cls, Module):
            raise TypeError("cls must be a subclass of torch.nn.Module")

        obj = cls.__new__(cls)
        obj.__init__(in_channels, out_channels, *args, **kwargs)
        return obj


class ModuleBase(Module, metaclass=ModuleMetaBase):
    """
    Basical abstract class of all neural network modules.
    Any subclass of this class must implement the following
    abstract methods:

    Parameters
    ----------
    in_channels : int
        The number of input channels.
    out_channels : int
        The number of output channels.
    **kwargs : Any
    """
    def __init__(self, in_channels: int, out_channels: int, **kwargs) -> None:
        super(ModuleBase, self).__init__()
        self._in_channels = in_channels
        self._out_channels = out_channels
        # store any other parameters required by the model
        self._kwargs = kwargs

    @abc.abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Data forward for a neural network waited to be implemented.

        Parameters
        ----------
        x : torch.Tensor
            The input data.

        Returns
        -------
        The output data.
        """
        raise NotImplementedError

    def parameters(self,
                   recurse: bool = True,
                   skip_no_grad: bool = False) -> Iterator[Parameter]:
        """
        Return model paramters with or without no grad parameters.
        If some layers are freezed, these parameters should not be
        updated by optimizer. Though no grad parameters won't be
        updated even if they are sent to optimizer. But it is a waste
        of calculation resource.

        Parameters
        ----------
        recurse : bool
            If True, return all parameters of all sublayers.
        skip_no_grad : bool, default False
            If True, return all parameters with grad.
             it advised set as True when using optimizer

        Returns
        -------
        An iterator of all parameters.
        """
        params = super().parameters(recurse=recurse)
        if skip_no_grad:
            return filter(lambda param: param.requires_grad, params)
        return params

    def freeze(self, layer: Module) -> None:
        """
        Freeze parameters of a layer and its sublayers

        Parameters
        ----------
        layer : Module
            The layer to be frozen.
        """
        for param in layer.parameters(recurse=True):
            param.requires_grad = False

    def unfreeze(self, layer: Module) -> None:
        """
        Unfreeze parameters of a layer and its sublayers

        Parameters
        ----------
        layer : Module
            The layer to be unfrozen.
        """
        for param in layer.parameters(recurse=True):
            param.requires_grad = True


class NetBase(ModuleBase):
    """
    Basic abstract class of all neural network models.
    Any subclass of this class must implement the following
    abstract methods:

    Parameters
    ----------
    in_channels : int
        The number of input channels.
    out_channels : int
        The number of output channels.
    shape : Tuple[int]
        The shape of a sample.
    **kwargs : Any
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 shape: Tuple[int] = None,
                 **kwargs) -> None:
        r"""
        Parameters:
        ----------
        - inChannels
        - outChannels
        - shape  the shape of a sample.
        """
        super().__init__(in_channels, out_channels, **kwargs)
        self._binary = False
        self._shape = shape

    @abc.abstractmethod
    def criterion(self, y_predict: torch.Tensor,
                  y_true: torch.Tensor) -> torch.Tensor:
        """
        Abstract methods to calculate the loss of the model.

        Parameters
        ----------
        y_predict : torch.Tensor
            The predicted data.
        y_true : torch.Tensor
            The true data.

        Returns
        -------
        The loss of the model.
        """
        raise NotImplementedError

    @property
    def shape(self) -> Tuple[int]:
        """
        Return the shape of a sample.

        Returns
        -------
        The shape of a sample.
        """
        if self._shape is None:
            raise TypeError("Uninitialized input tensor shape")
        return self._shape

    @shape.setter
    def shape(self, s: Tuple[int]):
        """
        Set the shape of a sample.

        Parameters
        ----------
        s : Tuple[int]
            The shape of a sample.

        Raises
        ------
        TypeError
            If the shape is not a tuple of int.
        """
        if len(s) <= 0:
            raise TypeError("Shape's length should greater than zero")

        for dim in s:
            if dim is not None and not isinstance(dim, int):
                raise TypeError("Invalid dimension")

        else:
            self._shape = s



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Tuple
import numpy as np
import pandas as pd
import sklearn.metrics as metrics

from socube.data import (onehot, plotAUC)

__all__ = ["getCurve", "evaluateReport", "binaryRate"]
"""
This module provides metric tools for training.
"""


def binaryRate(label: np.ndarray,
               predict: np.ndarray,
               positive_first: bool = False) -> Tuple[float]:
    """
    For binary classification task, calculate its confusion matrix
    and true positive rate (TPR), false negative rate (FNR), false
    positive rate (FPR) and true negative rate (TNR).

    Parameters
    ------------
    label: np.ndarray
        the true label 1D ndarray vector
    predict: np.ndarray
        the predict label 1D ndarray vector
    positive_first :If True, it will regard 0 as positive class. Otherwise
                it will regard 1 as positive class. By default, it is
                False.

    Returns
    ------------
    a tuple of TPR, FNR, FPR, TNR
    """
    confusion_matrix = metrics.confusion_matrix(label, predict)
    tnr, fpr, fnr, tpr = (confusion_matrix /
                          confusion_matrix.sum(axis=1).reshape(
                              (-1, 1))).flatten()
    if positive_first:
        tnr, fpr = tpr, fnr
    return tpr, fnr, fpr, tnr


def getCurve(label: np.ndarray,
             score: np.ndarray,
             average: str = "macro",
             curve: str = "roc"):
    """
    Evaluate receiver operating characteristic (ROC) for multiclass task
    based on false positive rate(FPR) and true positive rate(TPR). The area
    under curve(AUC) is used.

    Parameters
    ------------
    label: np.ndarray
        the true label 1D ndarray vector
    score: np.ndarray
        the predict score 1D ndarray vector
    average: str
        the average method, "macro" , "micro" or "binary"
        how to average multiclass. Because conventional
        ROC is designed for binary classification task. For
        detail about this, please search it on the Internet.

    Returns
    ------------
        A triple tuple of (x, y, AUC). When curve is "roc",
        x is FPR and y is TPR. When curve is "pr", x is FNR and y is TNR.
    """
    assert len(label) == len(score), "Unmatched label and score"
    assert average in ["macro", "micro",
                       "binary"], f"Invalid average mode {average}"

    def calc(y_true: np.ndarray, y_score: np.ndarray):
        if curve == "roc":
            x, y, _ = metrics.roc_curve(y_true, y_score)
        elif curve == "prc":
            y, x, _ = metrics.precision_recall_curve(y_true, y_score)
        else:
            raise NotImplementedError(f"Unsupport curve type '{curve}'")
        return x, y

    if average == "binary":
        assert len(label.shape) == 1, "Required 1D label for binary"
        assert len(score.shape) == 1, "Required 1D score for binary"
        x, y = calc(label, score)
    elif average == "macro":
        x_list = list()
        y_list = list()
        class_nums = score.shape[1]
        label = onehot(label, class_nums)
        for subtype in range(class_nums):
            sub_label = label[:, subtype]
            if not sub_label.max() > 0:
                continue
            sub_score = score[:, subtype]
            x_array, y_array = calc(sub_label, sub_score)
            x_list.append(x_array)
            y_list.append(y_array)
        x = np.unique(np.concatenate(x_list))
        y = np.zeros_like(x)
        for i in range(len(x_list)):
            y += np.interp(x, x_list[i], y_list[i])
        y /= i + 1
    elif average == "micro":
        label = onehot(label, score.shape[1])
        x, y = calc(label.ravel(), score.ravel())
    auc = metrics.auc(x, y)
    return x, y, auc


def evaluateReport(label: np.ndarray,
                   score: np.ndarray,
                   roc_plot_file: Optional[str] = None,
                   prc_plot_file: Optional[str] = None,
                   average: str = "macro",
                   threshold: float = 0.5) -> pd.Series:
    """
    Evaluate model performance with multiple indicator and generate file report.

    Parameters
    ------------
    label: np.ndarray
        the true label 1D ndarray vector
    score: np.ndarray
        the predict score 1D ndarray vector
    roc_plot_file: Optional[str]
        the file name of ROC curve plot. If None, it will not save plot.
    prc_plot_file: Optional[str]
        the file name of PR curve plot. If None, it will not save plot.
    average: str
        the average method, "macro" , "micro" or "binary"
        how to average multiclass. Because conventional
        ROC is designed for binary classification task. For
        detail about this, please search it on the Internet.
    threshold: float
        the threshold of predict score.

    Returns
    ------------
    a pandas.Series object as the report.
    """
    result = pd.Series(dtype="object")
    if len(score.shape) == 1:
        predict = (score >= threshold).astype(np.int64)
        average = "binary"
    elif len(score.shape) == 2:
        predict = score.argmax(1)
    else:
        raise NotImplementedError("Unsupport yscore dimension")

    report = metrics.classification_report(label, predict, output_dict=True)
    result["ACC"] = report["accuracy"]
    result["MCC"] = metrics.matthews_corrcoef(label, predict)
    result["AP"] = metrics.average_precision_score(label, score)

    roc_data_dict = dict()
    prc_data_dict = dict()

    fpr, tpr, auroc = getCurve(label, score, average, curve="roc")
    recall, precision, auprc = getCurve(label, score, average, curve="prc")
    roc_data_dict['%s ROC curve (area = %0.4f)' % (average, auroc)] = (fpr,
                                                                       tpr)
    prc_data_dict['%s PRC curve (area = %0.4f)' %
                  (average, auprc)] = (recall, precision)

    result[f"{average}_AUROC"] = auroc
    result[f"{average}_AUPRC"] = auprc

    result[f"{average}_F1"] = metrics.f1_score(label, predict, average=average)
    result[f"{average}_precision"] = metrics.precision_score(label,
                                                             predict,
                                                             average=average)
    result[f"{average}_recall"] = metrics.recall_score(label,
                                                       predict,
                                                       average=average)

    plotAUC(roc_data_dict, 'AUROC Curve', 'False Positive Rate',
            'True Positive Rate', roc_plot_file, 1)
    plotAUC(prc_data_dict, 'AUPRC Curve', 'Recall', 'Precision', prc_plot_file,
            -1)
    return result



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import warnings
import shutil
import numpy as np
import pandas as pd
import torch
import contextlib

from typing import (Callable, Optional, Dict, TypeVar, Union)
from pandas.core.generic import NDFrame

from .concurrence import parallel
from .context import ContextManagerBase
from .logging import log

__all__ = [
    "uniquePath", "mkDirs", "rm", "checkExist", "writeCsv", "writeNpy",
    "writeHdf", "loadTorchModule", "Redirect", "ReportManager"
]


def uniquePath(path: str, exists_warning: bool = False) -> str:
    """
    Check whether `path` existed in the file system. And generate a
    unique path by add index at the end of raw path.

    Parameters
    ----------------
    path : str
        The raw path to be checked.
    exists_warning : bool
        Whether to print warning when the path is existed.

    Returns
    ----------------
    str, the unique path.

    Examples
    ----------------
    >>> # pkgs/lapjv-1.3.1.tar.gz already existed
    >>> uniquePath("pkgs/lapjv-1.3.1.tar.gz")
    'pkgs/lapjv-1.3.1.tar(1).gz'

    >>> # pkgs/lapjv-1.3.1.tar.gz already existed and
    >>> # pkgs/lapjv-1.3.1.tar(1).gz already existed
    >>> uniquePath("pkgs/lapjv-1.3.1.tar.gz")
    'pkgs/lapjv-1.3.1.tar(2).gz'
    """
    name, ext = os.path.splitext(path)
    current_path = name
    idx = 1
    while os.path.exists(current_path + ext):
        current_path = "%s(%d)" % (name, idx)
        idx += 1

    if exists_warning and current_path != name:
        warnings.warn(
            f"File {name}{ext} already existed and will use {current_path}{ext} as instead!",
            RuntimeWarning,
            stacklevel=2)
    return current_path + ext


def mkDirs(path: str) -> None:
    r"""
    Create a directory if it not exists. If the parent is not existed,
    it will be created as well.

    Parameters
    ----------------
    path : str
        The path to be created.

    Examples
    ----------------
    >>> #parent is not existed
    >>> mkDirs("parent/target/")
    """
    if not os.path.exists(path):
        os.makedirs(path)


def rm(path: str) -> None:
    r"""
    Remove a directory or file.

    Parameters
    ----------------
    path : str
        The path to be removed.

    Examples
    ----------------
    >>> rm("targetFile")
    """
    if os.path.isdir(path):
        shutil.rmtree(path)
    if os.path.isfile(path):
        os.remove(path)


def checkExist(path: str,
               types: str = "file",
               raise_error: bool = True) -> bool:
    r"""
    Check whether a file or directory is existed.

    Parameters
    ---------------
    path : str
        The path to be checked.
    types : str
        The type of the path. "file" or "dir".
    raise_error : bool
        Whether to raise error when the path is not existed.

    Returns
    ---------------
    bool, whether the path is existed.

    Examples
    ---------------
    >>> checkExist("/public/file")
    >>> checkExist("/public/home/", types="dir")
    >>> checkExist("/public/home/", types="dir", raiseError=False)
    """
    if types == "file" and (not os.path.isfile(path)):
        if raise_error:
            raise FileNotFoundError(f"No such file {path}")
        return False
    elif types == "dir" and (not os.path.isdir(path)):
        if raise_error:
            raise FileNotFoundError(f"No such directory {path}")
        return False
    elif types not in ["file", "dir"]:
        raise TypeError(f"Invalid type {types}, only support file and dir")
    return True


@parallel
def writeCsv(data: Union[NDFrame, np.ndarray],
             file: str,
             callback: Optional[Callable] = None,
             **kwargs) -> None:
    """
    Write `ndarray` or any instance of `NDFrame` to a CSV file.

    Parameters
    ---------------
    data : Union[NDFrame, np.ndarray]
        The data to be written.
    file : str
        The path to be written.
    callback : Optional[Callable]
        The callback function to be called after writing.
    **kwargs
        The keyword arguments to be passed to `pandas.DataFrame.to_csv`.

    Examples
    ---------------
    >>> ndarray = np.array([1, 2, 3])
    >>> writeCsv(ndarray, "ndarray.csv", callback=lambda:print("Finish"))
    >>> series = pd.Series(ndarray)
    >>> writeCsv(series, "series.csv")
    """
    if isinstance(data, np.ndarray):
        data = pd.DataFrame(data)
        kwargs["index"] = None
        kwargs["header"] = None
    elif not isinstance(data, NDFrame):
        raise TypeError(f"Invalid data type {type(data)}")
    data.to_csv(file, quoting=2, **kwargs)
    if hasattr(callback, "__call__"):
        callback()


@parallel
def writeNpy(data: np.ndarray,
             file: str,
             callback: Optional[Callable] = None,
             allow_pickle: bool = True,
             **kwargs):
    """
    Write `ndarray` to NPY binary file. NPY format is
    numpy's specific file format.

    Parameters
    --------------
    data : np.ndarray
        The data to be written.
    file : str
        The path to be written.
    callback : Optional[Callable]
        The callback function to be called after writing.
    allow_pickle : bool
        Wether to allow use pickle to serialize python object.
    **kwargs
        The keyword arguments to be passed to `numpy.save`.

    Examples
    -------------
    >>> ndarray = np.array([1, 2, 3])
    >>> writeNpy(ndarray, "ndarray.npy")
    """
    np.save(file, data, allow_pickle=allow_pickle, **kwargs)
    if hasattr(callback, "__call__"):
        callback()


@parallel
def writeHdf(data: NDFrame,
             file: str,
             key="data",
             mode="w",
             callback: Optional[Callable] = None,
             **kwargs):
    r"""
    Write `NDFrame` instance to a HDF5 binary format file.

    Parameters
    -----------------
    data : NDFrame
        The data to be written.
    file : str
        The path to be written.
    key : str
        The key to be used to store the data.
    mode : str
        The mode to be used to open the file.
    callback : Optional[Callable]
        The callback function to be called after writing.
    **kwargs
        The keyword arguments to be passed to `pandas.DataFrame.to_hdf`.

    Examples
    ----------------
    >>> data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    >>> writeHdf(data, "data.hdf", callback=lambda:print("Finish"))
    """
    data.to_hdf(file, key=key, mode=mode, **kwargs)
    if hasattr(callback, "__call__"):
        callback()


def loadTorchModule(module: torch.nn.Module,
                    file: str,
                    skipped: bool = True,
                    verbose=False):
    """
    Load a torch module from a file.

    Parameters
    ----------------
    module : torch.nn.Module
        The module waited to be updated.
    file : str
        The module file path.
    skipped : bool
        Whether to skip the loading of the module when the file is not existed.
    verbose : bool
        Whether to print the loading message.
    """
    if os.path.isfile(file):
        module.load_state_dict(torch.load(file))
        if not verbose:
            log("loadTorchModule", f"model {file} loaded")
    elif skipped:
        log("loadTorchModule", f"File {file} not found and skipped")
    else:
        raise FileNotFoundError(f"File {file} not found")
    return module


# TODO fix the bug that logging module will ignore this Redirect.
class Redirect(ContextManagerBase):
    """
    A context manager to redirect stdout and stderr.

    Parameters
    ----------------
    target : str
        The redirect target file.
    verbose : bool
        Whether to print the redirect message.

    Examples
    ----------------
    >>> with Redirect("stdout.txt", verbose=True):
    >>>     print("Hello")
    >>>     print("World")
    """
    def __init__(self, target: str, verbose: bool = True) -> None:
        parent = os.path.dirname(target)
        mkDirs(parent)
        self.__target = open(target, "wt", encoding="utf-8")
        self.__verbose = verbose
        self.__stderr = contextlib.redirect_stderr(self.__target)
        self.__stdout = contextlib.redirect_stdout(self.__target)

    def __enter__(self):
        log(Redirect.__name__,
            f"Redirect output to {self.__target.name}",
            quiet=self.__verbose)

        self.__stdout.__enter__()
        self.__stderr.__enter__()
        return self

    def __exit__(self, *args) -> bool:
        self.__stderr.__exit__(*args)
        self.__stdout.__exit__(*args)
        self.__target.close()
        return False


ReportManagerType = TypeVar('ReportManagerType', bound="ReportManager")


class ReportManager(ContextManagerBase):
    """
    A context manager to save important data.

    Parameters
    ----------------
    reports : Dict[str, NDFrame]
        a dictionary of reports, key is report filename, value
        is a `NDFrame` object waited to be saved.
    verbose : bool
        Whether to print the saving message.
        If `True`, ReportManager will work in silence.

    Examples
    ----------------
    >>> with ReportManager({"report1": df1, "report2": df2}, verbose=True) as rm:
    >>>     print("Hello")
    >>>     print("World")
    >>>     rm.addReports("report1", df3)
    >>>     rm.addReports("report2", df4)
    """

    def __init__(self,
                 reports: Dict[str, NDFrame] = dict(),
                 verbose: bool = False) -> None:
        self.__reports = dict()
        self.__check(reports)
        self.__reports = reports
        self.__verbose = verbose

    def addReports(self, reports: Dict[str, NDFrame]) -> None:
        """
        add new reports in context

        Parameters
        ----------------
        reports : Dict[str, NDFrame]
            a dictionary of reports, key is report filename, value is a
            `NDFrame` object waited to be saved.
        """
        self.__check(reports)
        self.__reports.update(reports)

    def updateReportName(self, old: str, new: str) -> None:
        """
        Update the name of a report.

        Parameters
        ----------------
        old : str
            The old report name.
        new : str
            The new report name.
        """
        if old not in self.__reports:
            raise IndexError(f"Report {old} not added yet!")
        if not isinstance(new, str):
            raise TypeError("Report name expected a str object")
        self.__reports[new] = self.__reports.pop(old)

    def __log(self, msg: str, level: str = "info") -> None:
        """
        Log the message.
        """
        log(type(self).__name__, msg, level=level, quiet=self.__verbose)

    def __check(self, reports: Dict[str, NDFrame]) -> None:
        """
        Check the reports type.
        """
        for file, df in reports.items():
            assert isinstance(file,
                              str) and file != "", f"Invalid filename {file}"
            assert isinstance(
                df, NDFrame), "Report object should be series or dataframe"

    def __enter__(self) -> ReportManagerType:
        return self

    def __exit__(self, *args) -> bool:
        for file, df in self.__reports.items():
            try:
                writeCsv(df, file)
                self.__log(f"Report {file} saved!")
            except Exception as e:
                self.__log(
                    f"An error {e} occurred while trying to save file {file}",
                    "error")
        return False



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pandas as pd
import numpy as np
import abc
import os
import torch
from torch.utils.data import Dataset, Sampler, Subset
from sklearn.model_selection import StratifiedKFold

__all__ = ["DatasetBase", "ConvDatasetBase"]


class DatasetBase(Dataset, metaclass=abc.ABCMeta):
    """
    Abstract base class for datasets. All SoCube extended
    datasets must inherit and implement its abstract interface.

    Parameters
    ------------
    labels: pd.DataFrame
        Dataframe containing labels for each sample.
    shuffle: bool, default False
        Whether to shuffle each class's samples before splitting into batches.
        Note that the samples within each split will not be shuffled.
    seed: int, default None
        Random seed for shuffling.
    k: int, default 5
        Number of folds for k-fold cross-validation.
    task_type: str, default "classify"
        Type of task. Must be one of "classify", "regress".
    """
    def __init__(self,
                 labels: pd.DataFrame,
                 shuffle: bool = False,
                 seed: int = None,
                 k: int = 5,
                 task_type: str = "classify") -> None:

        if not shuffle:
            seed = None

        self._shuffle = shuffle
        self._seed = seed
        self._k = k
        self._labels = labels
        self._task_type = task_type

    def __len__(self) -> int:
        """Return dataset's size"""
        return self._labels.shape[0]

    @abc.abstractmethod
    def __getitem__(self, index: int) -> dict:
        """
        Abstract method to get the specified sample
        information according to the index

        Parameters
        ------------
        index: int
            Index of the sample

        Returns
        ------------
        sample: dict
            A dictionary containing the sample information,
            such as tensor data, label, etc.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sampler(self, subset: Subset) -> Sampler:
        """
        Abstract method for sampling a subset of this dataset.

        Parameters
        ------------
        subset: Subset
            A subset of this dataset.

        Returns
        ------------
        sampler: Sampler
            A sampler for the subset.
        """
        raise NotImplementedError

    @property
    def kFold(self):
        """
        Get generator for k-fold cross-validation dataset

        Returns
        ------------
        kFold: generator
            An generator for k-fold cross-validation dataset. Each iteration
            generates a tuple of two Subset objects for training and validating
        """
        skf = StratifiedKFold(n_splits=self._k,
                              random_state=self._seed,
                              shuffle=self._shuffle)

        # It will balance types if tasktype is classify
        if self._task_type == "regress":
            index = np.zeros((len(self), 1))
        elif self._task_type == "classify":
            index = self._labels
        else:
            raise ValueError("Task type must be one of 'classify', 'regress'")

        for train_index, valid_index in skf.split(index, index):
            yield (Subset(self, train_index), Subset(self, valid_index))


class ConvDatasetBase(DatasetBase):
    """
    Basical dataset designed for CNN.

    Parameters
    ----------------
    data_dir: str
        Path to the directory containing dataset.
    labels: pd.DataFrame
        Dataframe containing labels for each sample.
    transform: torch.nn.Module, default None
        Transform to apply to each sample.
    shuffle: bool, default False
        Whether to shuffle each class's samples before splitting into batches.
    seed: int, default None
        Random seed for shuffling.
    k: int, default 5
        Number of folds for k-fold cross-validation.
    task_type: str, default "classify"
        Type of task. Must be one of "classify", "regress".
    use_index: bool, default True
        Whether to use the numeric index as the sample file name, such as "0.npy",
        if `False`, then use the sample name in the labels as the sample file name,
        such as "sample_name.npy".
    """
    def __init__(self,
                 data_dir: str,
                 labels: pd.DataFrame,
                 transform: torch.nn.Module = None,
                 shuffle: bool = False,
                 seed: int = None,
                 k: int = 5,
                 task_type: str = "classify",
                 use_index: bool = True) -> None:
        super(ConvDatasetBase, self).__init__(labels, shuffle, seed,
                                              k, task_type)
        if not shuffle:
            seed = None
        self._data_dir = data_dir
        self._transform = transform
        self._use_index = use_index

    def __getitem__(self, index: int) -> dict:
        item: pd.Series = self._labels.iloc[index]
        if self._use_index:
            data_file = f"{index}.npy"
        else:
            data_file = f"{item.name}.npy"

        res = torch.from_numpy(
            np.load(os.path.join(self._data_dir, data_file)).astype("float32"))
        if self._transform:
            res = self._transform(res)
        return {'data': res, 'label': np.squeeze(item.values)}



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from concurrent.futures import Future
from typing import List, Optional
import sklearn.metrics as metrics

import torch.optim as optim
import torch.nn as nn
import torch
import numpy as np
import pandas as pd
import os
from socube.utils.concurrence import ParallelManager

from socube.utils.io import (checkExist, writeCsv, mkDirs, rm, loadTorchModule)
from socube.utils.memory import (GPUContextManager, visualBytes, autoClearIter)
from socube.utils.logging import (getJobId, log)
from socube.net import NetBase
from socube.train import (EarlyStopping, evaluateReport)
from .data import ConvClassifyDataset
from .model import SoCubeNet
from torch.utils.data import DataLoader, Subset, Dataset
from tqdm.auto import tqdm
"""A module provided for model training and evaluation"""
__all__ = ["fit", "validate", "infer"]


def fit(home_dir: str,
        data_dir: str,
        lr: float = 0.001,
        gamma: float = 0.99,
        epochs: int = 100,
        train_batch: int = 32,
        valid_batch: int = 500,
        transform: nn.Module = None,
        in_channels: int = 10,
        num_workers: int = 0,
        shuffle: bool = False,
        seed: int = None,
        label_file: str = "label.csv",
        threshold: float = 0.5,
        k: int = 5,
        once: bool = False,
        use_index: bool = True,
        gpu_ids: List[str] = None,
        step: int = 5,
        model_id: str = None,
        pretrain_model_path: str = None,
        max_acc_limit: float = 1,
        multi_process: bool = False,
        **kwargs) -> str:
    """
    Train socube model.

    Parameters
    ----------
    home_dir: str
        the home directory of the specfic job
    data_dir: str
        the dataset's directory
    lr: float
        learning rate, default: 0.001
    gamma: float
        learning rate decay, default: 0.99
    epochs: int
        training epochs, default: 100
    train_batch: int
        training batch size, default: 32
    valid_batch: int
        validation batch size, default: 500
    transform: nn.Module
        sample transform, such as `Resize`
    in_channels: int
        the number of input channels, default: 10
    num_workers: int
        the number of workers for data loading, default: 0
    shuffle: bool
        if `True`, data will be shuffled while k-fold cross-valid
    seed: int
        random seed for k-fold cross-valid or sample
    label_file: str
        the label file csv name, default: "label.csv",
    threshold: float
        the threshold for classification, default: 0.5
    k: int
        k value of k-fold cross-valid
    once: bool
        if `True`, k-fold cross-validation runs first fold only
    use_index: bool
        If `True`, it will read sample file by index. Otherwise, it will read
        sample file by sample name.
    device_name: str
        the device name, default: "cpu"
    step: int
        the epoch step of learning rate decay, default: 5
    model_id: str
        the model id, If `None`, it will be generated automatically
    pretrain_model_path: str
        the pretrain model path, if not `None`, it will load the pretrain model
    max_acc_limit: float
        the max accuracy limit, if the accuracy is higher than this limit, the
        training will stop to prevent overfitting.
    multi_process: bool
        if `True`, it will use multi-process to train the model.
    **kwargs: dict
        the other parameters wanted to be saved in the log file.

    Returns
    ----------
    job id string
    """
    if seed is not None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    model_id = model_id if model_id else getJobId()
    log("Train", "Model ID: " + model_id)

    dataset = ConvClassifyDataset(data_dir=data_dir,
                                  transform=transform,
                                  labels=label_file,
                                  seed=seed,
                                  shuffle=shuffle,
                                  k=k,
                                  use_index=use_index)
    shape = dataset[0]["data"].shape
    mkDirs(f"{home_dir}/outputs/{model_id}")
    mkDirs(f"{home_dir}/models/{model_id}")
    mkDirs(f"{home_dir}/plots/{model_id}")
    report = pd.DataFrame()
    train_record = pd.DataFrame()
    train_kwargs = {
        "dataset": dataset,
        "train_batch": train_batch,
        "valid_batch": valid_batch,
        "in_channels": in_channels,
        "num_workers": num_workers,
        "shape": shape,
        "gamma": gamma,
        "lr": lr,
        "epochs": epochs,
        "home_dir": home_dir,
        "model_id": model_id,
        "pretrain_model_path": pretrain_model_path,
        "step": step,
        "max_acc_limit": max_acc_limit,
        "threshold": threshold
    }
    kwargs.update({
        "model": SoCubeNet.__name__,
        "kFold": k,
        "transform": transform,
        "data_dir": data_dir,
        "seed": seed,
        "label": label_file,
        "onlyOne": once,
        "multiProcess": multi_process,
        "gpu_ids": gpu_ids
    })
    kwargs.update(train_kwargs)
    config = pd.Series(kwargs)

    # Specific the default gpu device by context manager
    # If not, pytorch with use cuda:0 as default and load data to it when intializing,
    # no matter what device you want to use.
    # If free memory of cuda:0 is less than required, a Runtime Error will occurred
    # Otherwise, you will find TWO gpu device are used by your process in `nvidia-smi` report.
    
    with ParallelManager(paral_type="process", max_workers=k if multi_process else 1, verbose=True) as pm:
        results: List[Future] = []
        for fold, (train_set, valid_set) in enumerate(dataset.kFold, 1):
            device = torch.device("cpu")
            if gpu_ids is not None and len(gpu_ids) > 0 and torch.cuda.is_available():
                device = torch.device(gpu_ids[(fold-1)%len(gpu_ids)])
            results.append(pm.submit(_fit, fold = fold, device=device, train_set=train_set, valid_set = valid_set,  **train_kwargs))
            # quit k-fold
            if once:
                break

        for result in results:
            rep, train_rep = result.result()
            for head in rep:
                report[head] = rep[head]
            for head in train_rep:
                train_record[head] = train_rep[head]

    report["average"] = report.mean(axis=1)
    report["sample_stdev"] = report.std(axis=1)
    writeCsv(
        pd.concat([report["average"], config]),
        f"{home_dir}/outputs/{model_id}/{SoCubeNet.__name__}_aveReport.csv",
        header=None)

    if epochs > 0:
        writeCsv(
            train_record,
            f"{home_dir}/outputs/{model_id}/{SoCubeNet.__name__}_trainRecord.csv",
            index=None)

    writeCsv(
        report,
        f"{home_dir}/outputs/{model_id}/{SoCubeNet.__name__}_kFoldReport.csv")

    return model_id


def _fit(
    device: torch.device,
    train_set: Subset,
    valid_set: Subset,
    dataset: Dataset,
    train_batch: int,
    valid_batch: int,
    in_channels: int,
    num_workers: int,
    shape: tuple,
    gamma: float,
    lr: float,
    fold: int,
    epochs: int,
    home_dir: str,
    model_id: str,
    step: int,
    max_acc_limit: float,
    threshold: float,
    pretrain_model_path: Optional[str] = None
):
    """
    A internal function to train one fold of k-fold cross-validation.
    """
    log("Train", f"Train Fold {fold}")
    with torch.cuda.device(device if device.type == "cuda" else -1), GPUContextManager():
        data_loader = DataLoader(dataset=train_set,
                                sampler=dataset.sampler(train_set),
                                batch_size=train_batch,
                                num_workers=num_workers)
        valid_loader = DataLoader(dataset=valid_set,
                                batch_size=valid_batch,
                                sampler=dataset.sampler(valid_set))

        # init and load pretrained model
        model = SoCubeNet(in_channels,
                        dataset.typeCounts,
                        freeze=pretrain_model_path,
                        binary=True,
                        shape=shape).to(device)

        log("Train", f"Shape is {model.shape}")
        if pretrain_model_path:
            mp = pretrain_model_path % (fold)
            log("Train", "Loading pretained model %s" % (mp))
            loadTorchModule(model, mp)

        optimizer = optim.Adam(model.parameters(), lr=lr)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer,
                                                        gamma=gamma)

        valid_loss, valid_acc, valid_label, valid_score = validate(
            data_loader=valid_loader, model=model, device=device)

        loop = tqdm(range(epochs))
        train_loss_array = np.full(epochs, -1, dtype=np.float32)
        valid_loss_array = np.full(epochs, -1, dtype=np.float32)
        train_acc_array = np.full(epochs, -1, dtype=np.float32)
        valid_acc_array = np.full(epochs, -1, dtype=np.float32)

        earlystop = EarlyStopping(
            path=f"{home_dir}/models/{model_id}/modelTmp_{fold}.pt",
            patience=10,
            verbose=10)

        for epoch in autoClearIter(loop):
            model.train()
            train_acc = 0
            train_loss = 0
            for index, batch in autoClearIter(enumerate(data_loader, 1)):
                data, label = batch.values()
                data = data.to(device)
                label = label.to(device)
                score = model(data)
                loss = model.criterion(score, label)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                if model._binary:
                    predict = (score >= 0.5).cpu().long()
                else:
                    predict = score.argmax(1).cpu()
                train_acc += metrics.accuracy_score(label.cpu(), predict)
                train_loss += loss.item()
                loop.set_description("Fold %02d batch %03d/%03d" %
                                    (fold, index, len(data_loader)))
                loop.set_postfix(
                    loss=train_loss / index,
                    train_acc=train_acc / index,
                    valid_acc=valid_acc,
                    best_loss=earlystop._best,
                    pid=os.getpid(),
                    memo_res=visualBytes(
                        torch.cuda.memory_reserved(device)),
                    max_memo_res=visualBytes(
                        torch.cuda.max_memory_reserved(device)),
                    lr="%.06f" % (scheduler.get_last_lr()[0]))

                del data, label, score, loss

            if (epoch + 1) % step == 0:
                scheduler.step()

            valid_loss, valid_acc, valid_label, valid_score = validate(
                data_loader=valid_loader, model=model, device=device)

            if earlystop(valid_loss, model, valid_acc < max_acc_limit):
                best_label = valid_label
                best_score = valid_score

            train_acc_array[epoch] = train_acc / index
            valid_acc_array[epoch] = valid_acc
            train_loss_array[epoch] = train_loss / index
            valid_loss_array[epoch] = valid_loss
            del valid_label, valid_score
            if earlystop.earlyStop:
                log(__name__, "Early stopped for training")
                break

        loadTorchModule(model, earlystop._path)
        model.cpu()
        torch.save(
            model.state_dict(),
            f"{home_dir}/models/{model_id}/{SoCubeNet.__name__}_{fold}.pt")
        rm(earlystop._path)
        evaluate_result = evaluateReport(
            best_label, best_score,
            f"{home_dir}/plots/{model_id}/{SoCubeNet.__name__}_ROC_{fold}.png",
            f"{home_dir}/plots/{model_id}/{SoCubeNet.__name__}_PRC_{fold}.png",
            threshold=threshold
        )

        log("Train", "Best validate ACC: %.4f" % (evaluate_result["ACC"]))
        del model
        result = [{f"fold_{fold}": evaluate_result}]
        if epochs > 0:
            result.append({
                f"fold_{fold}_train_loss": train_loss_array,
                f"fold_{fold}_valid_loss": valid_loss_array,
                f"fold_{fold}_train_ACC": train_acc_array,
                f"fold_{fold}_valid_ACC": valid_acc_array
            })
        return result


@torch.no_grad()
def validate(data_loader: DataLoader,
             model: NetBase,
             device: torch.device,
             with_progress: bool = False) -> tuple:
    """
    Validate model performance basically

    Parameters
    ----------
    dataLoader: the torch dataloader object used for validation
    model: Network model implemented `NetBase`
        the model waited for validation
    device: the cpu/gpu device

    Returns
    ----------
    a quadra tuple of (average loss, average ACC, true label, predict score)
    """
    with torch.cuda.device(device if device.type == "cuda" else -1):
        model.to(device)
        model.eval()
        loss_ave = 0
        acc_ave = 0
        score_list = list()
        label_list = list()

        itererate = autoClearIter(enumerate(data_loader, 1))
        if with_progress:
            itererate = tqdm(itererate, desc="Validate")

        for index, batch in itererate:
            data, label = batch.values()
            data = data.to(device)
            label = label.to(device)
            score = model(data)
            loss = model.criterion(score, label)
            if model._binary:
                predict = (score >= 0.5).cpu().long()
            else:
                predict = score.argmax(1).cpu()

            acc = metrics.accuracy_score(label.cpu(), predict)
            loss_ave += loss.cpu().item()
            acc_ave += acc
            score_list.extend(score.cpu().numpy())
            label_list.extend(label.cpu().numpy())

            # destroy useless object and recycle GPU memory
            del data, label, loss, score

        loss_ave /= index
        acc_ave /= index

    return loss_ave, acc_ave, np.array(label_list), np.array(score_list)


def infer(data_dir: str,
          home_dir: str,
          model_id: str,
          label_file: str,
          in_channels: int = 10,
          k: int = 5,
          threshold: float = 0.5,
          batch_size: int = 400,
          gpu_ids: List[str] = None,
          with_eval: bool = False,
          seed: Optional[int] = None,
          multi_process: bool = False):
    """
    Model inference

    Parameters
    ----------
    data_dir: str
        the directory of data
    home_dir: str
        the home directory of output
    model_id: str
        the id of model
    label_file: str
        the label file used to inference
    in_channels: int
        the number of input channels
    k: int
        k value for k-fold cross validation
    threshold: float
        the threshold for binary classification
    batch_size: int
        the batch size for inference
    gpu_ids: List[str]
        the list of gpu ids
    with_eval: bool
        whether to evaluate the model performance
    seed: int
        the seed for random
    multi_process: bool
        whether to use multi-process for inference
    """
    dataset = ConvClassifyDataset(data_dir=data_dir,
                                  labels=label_file,
                                  shuffle=False,
                                  seed=seed,
                                  use_index=False)
    
    log("Inference",
        f"Current threshold: {threshold}, batch_size: {batch_size}")
    model_dir = os.path.join(home_dir, "models", model_id)
    checkExist(model_dir, types="dir")
    output_dir = os.path.join(home_dir, "outputs", model_id)

    plot_dir = os.path.join(home_dir, "plots", model_id)
    if with_eval:
        mkDirs(plot_dir)
    mkDirs(output_dir)
    ensemble_score_list = []
    with ParallelManager(max_workers=k if multi_process else 1, paral_type="process", verbose=True) as pm:
        results: List[Future] = []
        for fold in range(1, k + 1):
            device = torch.device("cpu")
            if gpu_ids is not None and len(gpu_ids) > 0 and torch.cuda.is_available():
                device = torch.device(gpu_ids[(fold-1)%len(gpu_ids)])
            results.append(pm.submit(_infer,
                in_channels,
                model_dir,
                dataset,
                batch_size,
                device,
                with_eval,
                plot_dir,
                threshold,
                fold,
                output_dir))
        for result in results:
            ensemble_score_list.append(result.result())
    log("Inference", "Model ensembling")
    ensemble_score = sum(ensemble_score_list) / k

    if with_eval:
        writeCsv(
            evaluateReport(
                label=dataset._labels.iloc[:, 0].values,
                score=ensemble_score,
                roc_plot_file=os.path.join(plot_dir, f"inference_roc_{threshold}.png"),
                prc_plot_file=os.path.join(plot_dir, f"inference_prc_{threshold}.png"),
                threshold=threshold),
            os.path.join(output_dir, f"inference_report_{threshold}.csv"))

    result = pd.DataFrame({"predict_score": ensemble_score},
                          index=dataset._labels.index)
    result["predict_type"] = result["predict_score"].apply(
        lambda s: "doublet" if s >= threshold else "singlet")
    writeCsv(result, os.path.join(output_dir, f"final_result_{threshold}.csv"))


def _infer(
    in_channels,
    model_dir,
    dataset,
    batch_size,
    device,
    with_eval,
    plot_dir,
    threshold,
    fold,
    output_dir):
    """
    Internal function for inference
    """
    model = SoCubeNet(in_channels=in_channels, out_channels=2, binary=True)
    dataloader = DataLoader(dataset=dataset, batch_size=batch_size)
    loadTorchModule(model,
                    os.path.join(model_dir, f"{SoCubeNet.__name__}_{fold}.pt"),
                    skipped=False)
    _, _, label, score = validate(dataloader, model, device, True)
    assert (dataset._labels.iloc[:, 0] == label).all()
    if with_eval:
        writeCsv(
            evaluateReport(
                label=label,
                score=score,
                roc_plot_file=os.path.join(plot_dir,
                                    f"inference_roc_{fold}_{threshold}.png"),
                prc_plot_file=os.path.join(plot_dir,
                                    f"inference_prc_{fold}_{threshold}.png"),
                threshold=threshold),
            os.path.join(output_dir, f"inference_report_{fold}_{threshold}.csv"))

    writeCsv(pd.DataFrame({"score": score}, index=dataset._labels.index),
                os.path.join(output_dir, f"inference_score_{fold}.csv"))
    return score



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import multiprocessing as mp
from concurrent.futures import (Executor, ThreadPoolExecutor,
                                ProcessPoolExecutor, Future)
from typing import Callable, List, TypeVar
from .context import ContextManagerBase
from .logging import log

import functools
"""
This module is developed for concurrence programing.
"""
__all__ = ["parallel", "ParallelManager"]

GLOBAL_POOL_STACK: List[Executor] = []
GLOBAL_TASK_STACK: List[List[Future]] = []


def parallel(func):
    """
    A decorator for asynchronous method

    Parameters
    ----------
    func : Callable
        The function to be decorated as asynchronous.

    Returns
    ----------
        a new function. If it is called under ParallelManager, it will be
        executed in a process pool.

    Examples
    ----------
    >>> @parallel
        def func(*args, **kwargs):
            ...
    >>> parallelFunc = parallel(func)
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # add task to nearest ThreadPoolExecutor
        if len(GLOBAL_TASK_STACK) > 0 and len(GLOBAL_POOL_STACK) > 0:
            for pool_index in range(len(GLOBAL_POOL_STACK)):
                if isinstance(GLOBAL_POOL_STACK[-pool_index-1], ThreadPoolExecutor):
                    task = GLOBAL_POOL_STACK[-pool_index-1].submit(func, *args, **kwargs)
                    GLOBAL_TASK_STACK[-pool_index-1].append(task)
                    return task
            else:
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return wrapper


ParallelManagerType = TypeVar("ParallelManagerType", bound="ParallelManager")


class ParallelManager(ContextManagerBase):
    """
    A context manager for multi-thread. In this context, Multiple thread is enable
    will be set as True and enable multi-thread for function decorated with `parallel`.
    Multi-level nesting is allowed, and each level has an independent asynchronous
    task pool. When exiting the context, it will wait for the tasks in the current
    asynchronous task pool to complete.

    Parameters
    ----------
    verbose : bool
        Whether to print log.
    max_workers : int
        The maximum number of workers in the asynchronous task pool.
    paral_type : str
        The type of asynchronous task pool. Only support "thread" and "process".
    raise_error : bool
        Whether to raise error when the task is not completed. Default is True.
        If you do not want interrupt the program, set it to False.

    Examples
    ----------
    >>> with ThreadManager() as tm:
            # Here is you code
            asynFunc1(*args)  # this function is decorated with `parallel`
            with tm:
                asynFunc2(*args)  # this function is decorated with `parallel`
            print("asynFunc2 finished") # this code will not be executed util asynFunc2 finished.
        print("asynFunc1 finished")  # this code will not be executed util asynFunc1finished.
    """

    def __init__(self,
                 verbose: bool = False,
                 max_workers: int = 8,
                 paral_type: str = "thread",
                 raise_error: bool = True) -> None:
        self.verbose = verbose
        self.max_workers = max_workers
        self.paral_type = paral_type
        self.pool = None
        self.raise_error = raise_error
        self.task_stack: List[Future] = []
        if paral_type not in ["process", "thread"]:
            raise TypeError(
                f"Unsupport paral_type {paral_type}, only support process/thread."
            )

    def __enter__(self) -> ParallelManagerType:
        if self.paral_type == "process":
            pool = ProcessPoolExecutor(self.max_workers, mp_context=mp.get_context("spawn"))
        elif self.paral_type == "thread":
            pool = ThreadPoolExecutor(self.max_workers)

        GLOBAL_POOL_STACK.append(pool)
        GLOBAL_TASK_STACK.append(self.task_stack)
        self.pool = pool
        log(type(self).__name__,
            f"Enter a multi-{self.paral_type} context",
            quiet=self.verbose)
        return self

    def submit(self, func: Callable, *args, **kwargs)->Future:
        """
        Add a task to the asynchronous task pool.
        For muti-thread task, you can use `parallel` decorator instead.
        But it not support multi-process task.
        """
        task = self.pool.submit(func, *args, **kwargs)
        self.task_stack.append(task)
        return task

    def __exit__(self, *args) -> bool:
        assert GLOBAL_POOL_STACK.pop() is self.pool
        assert GLOBAL_TASK_STACK.pop() is self.task_stack
        for task in self.task_stack:
            ex = task.exception()
            if ex is not None:
                if self.raise_error:
                    raise ex
                else:
                    log(type(self).__name__, f"Sub {self.paral_type} task {task} exited with error: {ex}")
        self.pool.shutdown(wait=True)
        log(type(self).__name__,
            f"Quit a multi-{self.paral_type} context",
            quiet=self.verbose)
        return False


# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from types import TracebackType
from typing import Type, TypeVar

from .context import ContextManagerBase
from .mail import MailService
from .logging import log

import traceback
import signal

__all__ = ["ExceptionManager"]

ExceptionManagerType = TypeVar('ExceptionManagerType',
                               bound="ExceptionManager")


class ExceptionManager(ContextManagerBase):
    """
    A context manager to deal with exception and notify users if it is required.
    Any exception occurred within this context will be catched and send notification
    email to user if mailService is not equal to 'None'. You use it as following.

    Parameters
    ----------
    mail_service : MailService
        A MailService object used to send email to users, If None, no email will be sent
    is_continue : bool
        If continue to run this program when an error occurred.
    ignore_keyboard_interrupt : bool
        If true, it will ignore keyboard interrupt signal, such as "Ctrl + C". It prevent
        quit program by pressing keyboard unexpectedly. But it also is a shortcoming
        when you want to terminate program.

    Examples
    ----------
    >>> mailService = MailService(...)
    >>> with ExceptionManager(mailService)
    >>>     a = 1/0  # will throw ZeroDivideError
    """

    def __init__(self,
                 mail_service: MailService = None,
                 is_continue: bool = False,
                 ignore_keyboard_interrupt: bool = False,
                 ignore_normal_system_exit: bool = True) -> None:
        self._mail_service = mail_service
        self._continue = is_continue
        self._ignore_keyboard_interrupt = ignore_keyboard_interrupt
        self._ignore_normal_system_exit = ignore_normal_system_exit
        self._normal_text = None

    def setNormalInfo(self, text: str) -> None:
        """
        Set the notification email when quiting context without any exception occurred.
        This can be used to send a processing result. Result should as short as possible.

        Args:
            text         email main text

        >>> with ExceptionManager(mailService) as em:
        >>>     # Here are some processing codes
        >>>     em.setNormalInfo("Prcoessing results")
        """
        self._normal_text = text

    def setMailService(self, mail_service: MailService) -> None:
        """
        Set the mail service object.
        """
        self._mail_service = mail_service

    def __enter__(self: ExceptionManagerType) -> ExceptionManagerType:
        if self._ignore_keyboard_interrupt:
            signal.signal(
                signal.SIGINT,
                lambda s, f: log("Exception", "Ignore sginal {s}: Ctrl + C"))
        return self

    def __exit__(self, exc_type: Type[Exception], exc_obj: Exception,
                 exc_trace: TracebackType) -> bool:
        try:
            assert exc_obj is None or isinstance(
                exc_obj, exc_type), "excVal must be a instance of excType!"

            if self._ignore_keyboard_interrupt:
                signal.signal(signal.SIGINT, signal.default_int_handler)

            skipped = self._ignore_normal_system_exit and isinstance(
                exc_obj, SystemExit) and exc_obj.code == 0

            if isinstance(self._mail_service, MailService):
                with self._mail_service as mail:
                    if skipped:
                        log(
                            "Exception",
                            "Ignore sginal SystemExit: exit code {code}".
                            format(code=exc_obj.code))
                    elif exc_type is not None:
                        mail.sendSelf(
                            "Program notification", "".join(
                                traceback.format_exception(
                                    exc_type, exc_obj, exc_trace, 1)))
                    elif self._normal_text is not None:
                        mail.sendSelf("Program notification",
                                      self._normal_text)
            # if not continue, error will raise again so it is not necessary to print it ourself
            if self._continue and exc_obj and not skipped:
                traceback.print_exception(exc_type, exc_obj, exc_trace)
        except Exception as e:
            log("Exception",
                f"A unexpected error occured while exiting context:{e}")
        finally:
            return self._continue



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Tuple
import torch
import os
import numpy as np
import pandas as pd

from pandas.core.generic import NDFrame
from torch.utils.data import (WeightedRandomSampler, Subset)
from socube.data.loading import ConvDatasetBase
from socube.utils.concurrence import parallel
from socube.utils.io import uniquePath, writeCsv, writeHdf
from socube.utils.logging import log

__all__ = ["ConvClassifyDataset", "generateDoublet", "checkShape", "checkData", "createTrainData"]


class ConvClassifyDataset(ConvDatasetBase):
    """
    Class of dataset for socube

    Parameters
    ------------
    data_dir: string
        the dataset's directory
    transform: torch module
        sample transform, such as `Resize`
    labels: string
        the label file csv name
    shuffle: boolean value
        if `True`, data will be shuffled while k-fold cross-valid
    seed: random seed
        random seed for k-fold cross-valid or sample
    k: integer scalar value
        k value of k-fold cross-valid
    use_index: boolean value
        If `True`, it will read sample file by index.
    """

    def __init__(self,
                 data_dir: str,
                 transform: torch.nn.Module = None,
                 labels: str = "label.csv",
                 shuffle: bool = False,
                 seed: int = None,
                 k: int = 5,
                 use_index: bool = True) -> None:

        super(ConvClassifyDataset,
              self).__init__(data_dir,
                             pd.read_csv(os.path.join(data_dir, labels),
                                         index_col=0,
                                         header=None),
                             transform=transform,
                             shuffle=shuffle,
                             seed=seed,
                             k=k,
                             use_index=use_index,
                             task_type="classify")

    def sampler(self, subset: Subset) -> WeightedRandomSampler:
        """
        Generate weighted random sampler for a subset of this dataset

        Parameters
        ------------
        subset: Subset
            the subset of this dataset

        Returns
        ------------
        Weighted random sampler
        """
        assert subset.dataset == self, "Must be a subset of this dataset"
        labels = self._labels[1][subset.indices]
        numSamples = labels.shape[0]
        labelWeights = numSamples / np.bincount(labels)
        sampleWeights = labelWeights[labels]
        generator = torch.Generator().manual_seed(
            self._seed) if self._seed is not None else None

        return WeightedRandomSampler(sampleWeights,
                                     numSamples,
                                     generator=generator)

    @property
    def typeCounts(self) -> int:
        """Numbers of different types"""
        return self._labels[1].unique().shape[0]


def generateDoublet(samples: pd.DataFrame,
                     ratio: float = 1.0,
                     adj: float = 1.0,
                     seed: Optional[int] = None,
                     size: Optional[int] = None) -> Tuple[pd.DataFrame]:
    """
    Generate training set from samples. in silico doublet
    will be simulated as positive samples.

    Parameters
    ------------
    samples: pd.DataFrame
        the samples dataframe, with row as cells (droplets, simples)
        and column as genes.
    ratio: float, default 1.0
        The ratio of the number of doublet and singlet.
    adj: float, default 1.0
        The adjustment factor for the doublet expression level. Generally,
        doublet is considered to have twice the gene expression level of
        singlet, but this is not necessarily the case in some cases. The gene
        expression level of the generated doublet is adjusted by the
        adjustment factor.
    seed: int, default None
        The random seed for the generation of the doublet.
    size: int, default None
        The size of the generated training set. If `None`, the size of the
        training set will be the same as the size of the samples.

    Returns
    ------------
    a tuple of two pd.DataFrame, the first is the positive (doublet) samples,
    the second is the negative (singlet) samples.
    """
    values = samples.values
    droplet_num = samples.shape[0]
    if size is None or size <= 0:
        size = droplet_num
    doublet_num = int(ratio * size / (ratio + 1))
    random = np.random.RandomState(seed)
    pair_index = random.choice(droplet_num, size=(doublet_num, 2))
    doublets = pd.DataFrame(values[pair_index[:, 0]] +
                            values[pair_index[:, 1]] * adj,
                            columns=samples.columns,
                            index=[f"doublet_{i}" for i in range(doublet_num)])

    singlets = samples.sample(size - doublet_num, random_state=seed)
    return singlets, doublets


def checkShape(path: str, shape: tuple = (10, None, None)) -> None:
    """
    Check dataset shape

    Parameters
    ------------
    path: string
        the dataset's directory
    shape: tuple, default (10, None, None)
        the expected shape of the dataset, None means any shape
    
    Raises
    ------------
    AssertionError: if the shape of the dataset is not the same as the expected
    """
    for file in filter(lambda x: x.endswith(".npy"), os.listdir(path)):
        test_sample: np.ndarray = np.load(os.path.join(path, file))
        assert len(test_sample.shape) == len(
            shape
        ), f"The shape of dataset {path} is not satisfied with {shape}"
        for true, expect in zip(test_sample.shape, shape):
            assert expect is None or true == expect, f"The shape of dataset {path} is not satisfied with {shape}"
        break


def checkData(data: pd.DataFrame):
    """
    Data legitimacy verification

    Parameters
    ----------
    data : pd.DataFrame
        Data to be checked, a dataframe of scRNA-seq data

    Raises
    ------
    ValueError
        If data contains NaN or inf
     IndexError
        If data contains duplicate column or row names,
        or if droplet name begins with "doublet"
    """
    def _check(x: pd.Series):
        if x.name.startswith("doublet"):
            raise IndexError(
                f"Droplet's name started with 'doublet' is not allowed, please rename it. Error in droplet: {x.name}"
            )
        elif x.isna().any():
            raise ValueError(f"A 'nan' value found at droplet {x.name}")

    if not data.index.is_unique:
        raise IndexError(
            "Duplicated droplet name detected, but name uniqueness is required"
        )

    if not data.columns.is_unique:
        raise IndexError(
            "Duplicated gene name detected, but name uniqueness is required")

    data.apply(_check, axis=1)


@parallel
def createTrainData(samples: pd.DataFrame,
                    output_path: str,
                    ratio: float = 1,
                    adj: float = 1,
                    seed: Optional[int] = None) -> Tuple[NDFrame]:
    """
    Based on the original data, doublets are generated as
    the positive data and a subset of the original data is used
    as the negative data to obtain the training dataset.

    Parameters
    ----------
    samples : pd.DataFrame
        Original data, a dataframe of scRNA-seq data. Shape is (n_droplets, n_genes)
    output_path : str
        Path to save the generated training data.
    ratio: float, default 1.0
        The ratio of the number of doublet and singlet.
    adj: float, default 1.0
        The adjustment factor for the doublet expression level. Generally,
        doublet is considered to have twice the gene expression level of
        singlet, but this is not necessarily the case in some cases. The gene
        expression level of the generated doublet is adjusted by the
        adjustment factor.
    seed: int, default None
        The random seed for the generation of the doublet.

    Returns
    -------
    A tuple of NDFrames. The first element is training data, the second element
    is the training label.
    """
    log("Generate", f"Generating doublet with ratio {ratio} and adj {adj}...")
    singlets, doublets = generateDoublet(samples,
                                          seed=seed,
                                          ratio=ratio,
                                          adj=adj,
                                          size=len(samples))
    train_data = pd.concat([samples, doublets])

    log("Generate", "Writing dataset...")
    writeHdf(
        train_data,
        uniquePath(
            os.path.join(
                output_path,
                f"02-trainData[{train_data.dtypes.iloc[0].name}][raw+samples({ratio},{adj})].h5"
            ), True))

    negative = pd.Series(np.zeros_like(singlets.index, dtype=np.int8),
                         index=singlets.index)
    positive = pd.Series(np.ones_like(doublets.index, dtype=np.int8),
                         index=doublets.index)
    train_label = pd.concat([negative, positive])
    writeCsv(train_label,
             uniquePath(
                 os.path.join(output_path, f"label[raw+samples({ratio},{adj})].csv"),
                 True),
             header=None)
    log("Generate", "Train data created")
    return train_data, train_label



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import torch
import torch.nn as nn

from socube.net import (NetBase, InceptionBlock2DLayer)

__all__ = ["SoCubeNet"]


class SoCubeNet(NetBase):
    """
    Neural network model constructed for doublet detection task. previous name
    is Conv2DClassifyNet.

    Parameters
    ----------
    in_channels: integer scalar value
        input data's channel count
    out_channels: integer scalar value
        output data's channel count
    freeze: boolean value
        Whether to freeze the feature extraction layer, default is `False`
    binary: boolean value
        Whether the output probability is binary or multicategorical, default is `True` for binary

    Examples
    ----------
    >>>  SoCubeNet(10, 2)
    """
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 freeze: bool = False,
                 binary: bool = True,
                 **kwargs):
        super(SoCubeNet, self).__init__(in_channels, out_channels, **kwargs)
        self._feature = InceptionBlock2DLayer(in_channels, 192)
        self._binary = binary and (out_channels == 2)
        if self._binary:
            out_channels = 1
        if freeze:
            self.freeze(self._feature)

        self._classify = nn.Sequential(
            nn.Linear(192, 128),
            nn.ReLU(inplace=True),
            nn.Linear(128, 32),
            nn.ReLU(inplace=True),
            nn.Linear(32, out_channels))
        if self._binary:
            self._classify.add_module(str(len(self._classify)), nn.Flatten(0))
            self._classify.add_module(str(len(self._classify)), nn.Sigmoid())
        else:
            self._classify.add_module(str(len(self._classify)), nn.Softmax(1))

    def forward(self, x1: torch.Tensor) -> torch.Tensor:
        hidden = self._feature.forward(x1)
        return self._classify(hidden)

    def criterion(self, yPredict: torch.Tensor,
                  yTrue: torch.Tensor) -> torch.Tensor:
        if self._binary:
            predictLoss = nn.BCELoss()(yPredict, yTrue.float())
        else:
            predictLoss = nn.CrossEntropyLoss()(yPredict, yTrue)
        return predictLoss



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from email.generator import Generator
import torch.cuda as cuda

from typing import Any, Iterable, List, TypeVar
from .context import ContextManagerBase
from .logging import log

"""
This module provides some tools for GPUs/CPUs Memory
"""

__all__ = ["visualBytes", "getGPUReport", "autoClearIter"]


def visualBytes(size: int) -> str:
    """
    Make memory size more friendly with proper unit.

    Parameters
    ----------
    size: a int type of bytes szie

    Returns
    ----------
    a formated string with size unit.

    Examples
    ----------
    >>> visualBytes(2<<20)
    '2.00MB'
    >>> visualBytes(18200565665)
    '16.95GB'
    """
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    idx = 0
    while size >= (1 << 10) and idx + 1 < len(units):
        size /= (1 << 10)
        idx += 1
    return "%.2f%s" % (size, units[idx])


def getGPUReport(*devices: int) -> str:
    """
    Get gpu memory usage of this program by pytorch.

    Parameters
    ----------
    devices: a list of device id

    Returns
    ----------
    A string report of gpu memory
    """
    if len(devices) == 0:
        devices = range(cuda.device_count())
    reports = map(
        lambda idx:
        f"{idx}\t cached: {visualBytes(cuda.memory_reserved(idx))}", devices)
    return "\n".join(reports)


GPUCacheManagerType = TypeVar("GPUCacheManagerType", bound="GPUContextManager")


class GPUContextManager(ContextManagerBase):
    """
    A context manager to help user automatically clean cached GPU memory.

    Parameters
    ----------
    quiet: boolean value
        Whether to remain silent

    Examples
    ----------
    >>> with GPUContextManager():
    >>>     # code using gpu by torch
    >>>     data = data.to(torch.device(0))
    >>>     pred = model(data)
    """
    def __init__(self, quiet: bool = True) -> None:
        self.quiet = quiet

    def __enter__(self) -> GPUCacheManagerType:
        cuda.empty_cache()
        return self

    def __exit__(self, *args) -> bool:
        cuda.empty_cache()
        log(GPUCacheManagerType.__name__,
            "\n" + getGPUReport(),
            quiet=self.quiet)
        return False


def autoClearIter(iter: Iterable[Any]) -> Generator:
    """
    Create a generator to automatically clean GPU memory when iterating.
    """
    for item in iter:
        with GPUContextManager():
            yield item


def parseGPUs(gpu_str: str) -> List[int]:
    """
    Parse a string of gpu ids.

    Parameters
    ----------
    gpu_str: a string of gpu ids.

    Returns
    ----------
    a list of gpu ids.

    Examples
    ----------
    >>> parseGPUs("0,1,2")
    [0, 1, 2]
    >>> parseGPUs("0")
    [0]
    >>> parseGPUs("")
    []
    """
    try:
        if gpu_str == "":
            return []
        return ["cuda:%d"%(int(idx)) for idx in gpu_str.split(",")]
    except Exception as e:
        raise ValueError(f"Invalid gpu id: {gpu_str}") from e



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__version__ = "0.1rc2"
__author__ = "Zhang.H.N"
__email__ = "zhang.h.n@foxmail.com"
__url__ = "https://github/idrblab/socube/"


def main(*args: str):
    from argparse import ArgumentParser
    parser = ArgumentParser("SoCube")

    parser.add_argument("--version",
                        "-v",
                        action="store_true",
                        help="Print socube version info")
    basic_args = parser.add_argument_group("basic configuration")
    basic_args.add_argument("--input",
                            "-i",
                            type=str,
                            default=None,
                            help='''
                        UMI expression matrix for scRNA-seq, supporting h5 format
                        (pandas DataFrame) and h5ad format (implemented in python
                        package anndata). The rows of the data are cells and the columns
                        are genes, please make sure that the cell ID is unique.
                        ''')
    basic_args.add_argument("--output",
                            "-o",
                            default=None,
                            type=str,
                            help='''
                        The output directory of the processed intermediate files and the
                        final result, if not specified, the default is the same directory as the
                        input file, if the output directory does not exist, a new one will be
                        created, if it already exists, continue to use the existing directory.
                        There are subdirectories embedding for socube embedding features,
                        models, plots and outputs for models, graphs and results generated
                        by model training. The final results are in the outputs subdirectory.
                        ''')
    basic_args.add_argument("--gpu-ids",
                            type=str,
                            default=None,
                            help='''
                        Ids of GPUs for accelerated computing, such as "0,1,2,3"
                        ''')
    basic_args.add_argument("--seed",
                            type=int,
                            default=None,
                            help="Random seed to reproduce, default is not specific")
    basic_args.add_argument("--k",
                            type=int,
                            default=5,
                            help='''
                        The k-fold cross-validation is used in training, and the resulting k models
                        will be ensembled into one model. Default 5.
                        ''')
    basic_args.add_argument("--adj-factor",
                            type=float,
                            default=1.0,
                            help='''
                        The adjustment factor for the doublet expression level. By default it is
                        assumed that the doublet expression level is twice the sinlget, but there
                        are fluctuations in the real situation and the expression level can be
                        changed by adjusting this factor. Default 1.0.
                        ''')
    basic_args.add_argument("--dim",
                            "-d",
                            type=int,
                            default=10,
                            help='''
                        The target dimension for gene degradation is also the number of
                        channels to train the model. Default 10.
                        ''')
    basic_args.add_argument("--cube-id",
                            type=str,
                            default=None,
                            help='''
                        If you want to reuse the socube embedding features obtained earlier,  just
                        specify the embedding ID, which is a string like \"yyyymmdd-HHMMSS-xxx\",
                        along with the original output path.
                        ''')

    model_args = parser.add_argument_group("model training configuration")
    model_args.add_argument("--learning-rate",
                            "-lr",
                            type=float,
                            default=0.001,
                            help="Learning rate of model training. Default 1e-3.")
    model_args.add_argument("--epochs",
                            "-e",
                            type=int,
                            default=100,
                            help="Epochs of model training. Default 100")
    model_args.add_argument("--train-batch-size",
                            type=int,
                            default=64,
                            help="Batch size of model training. Default 64.")
    model_args.add_argument("--vaild-batch-size",
                            type=int,
                            default=512,
                            help="Batch size of model validating. Default 512.")
    model_args.add_argument("--infer-batch-size",
                            type=int,
                            default=400,
                            help="Batch size of model inferring. Default 400.")
    model_args.add_argument("--threshold",
                            "-t",
                            type=float,
                            default=0.5,
                            help='''
                            The classification threshold for doublet detection. The
                            model outputs the probability score of doublet, which
                            is greater than the threshold considered as doublet and
                            vice versa for singlet. user can customize the threshold value.
                            Default 0.5.
                        ''')
    model_args.add_argument("--enable-validation",
                            "-ev",
                            action="store_true",
                            help='''
                        This optional is provided for performance evaluation.
                        You should input h5ad format data, and store label in
                        its `obs` property. `obs` property is a `DataFrame` object
                        and its label column named "type" and value is "doublet"
                        and "singlet".
                        ''')
    model_args.add_argument("--enable-multiprocess",
                            "-mp",
                            action="store_true",
                            help='''
                            Enable multi process to make use of CPU's multiple cores.
                            ''')

    notice_args = parser.add_argument_group("notice configuration")
    notice_args.add_argument(
        "--mail",
        type=str,
        default=None,
        help="Email address to send and receive the notification.")
    notice_args.add_argument(
        "--mail-server",
        type=str,
        default=None,
        help="Email server address. such as smtp.gmail.com")
    notice_args.add_argument("--mail-port",
                             type=int,
                             default=465,
                             help="Email server port, such as 994 for SSL.")
    notice_args.add_argument("--mail-passwd",
                             type=str,
                             default=None,
                             help="Email account password")
    notice_args.add_argument("--enable-ssl",
                             action="store_true",
                             help="Enable SSL for mail service")
    args = parser.parse_args(args if args else None)

    if args.version:
        print("Release v{:s}, Copyright (c) 2022 {:s}".format(
            __version__, __author__))
        return

    if args.input is None and args.cube_id is None:
        parser.print_help()
        return

    # Avoid import libraries at head
    from socube.utils.logging import getJobId, log
    log("Config", "Load required modules")
    import shutil
    import numpy as np
    import scanpy as sc
    import pandas as pd
    import torch

    from concurrent.futures import Future
    from socube.task.doublet import (createTrainData, checkData)
    from socube.data.preprocess import minmax, std
    from socube.utils.exception import ExceptionManager
    from socube.utils.mail import MailService
    from socube.utils.memory import parseGPUs
    from socube.cube import SoCube
    from socube.utils.io import mkDirs, writeCsv, writeHdf, checkExist
    from socube.utils.concurrence import ParallelManager
    from socube.data import filterData
    from socube.task.doublet import fit, infer, checkShape
    from os.path import dirname, join
    from scipy.sparse import issparse

    pd.options.mode.use_inf_as_na = True

    if None not in [
            args.mail, args.mail_server, args.mail_port, args.mail_passwd
    ]:
        log("Config", "Mail service inited for notification")
        mail_service = MailService(sender=args.mail,
                                   passwd=args.mail_passwd,
                                   server=args.mail_server,
                                   port=args.mail_port,
                                   ssl=args.enable_ssl)
    else:
        mail_service = None

    with ExceptionManager(mail_service=mail_service) as em:
        data_path = dirname(args.input) if args.input else "."
        home_path = args.output if args.output is not None else data_path
        cube_id = args.cube_id if args.cube_id else getJobId()
        embedding_path = join(home_path, "embedding", cube_id)
        train_path = join(embedding_path, "traindata")
        my_cube = SoCube.instance(map_type="pca",
                                  data_path=data_path,
                                  home_path=home_path)
        gpu_ids = None
        if torch.cuda.is_available():
            if args.gpu_ids is None:
                log("Config", "Use CPU for training, but GPU is available, specify '--gpu-ids' to use GPU")
            else:
                gpu_ids = parseGPUs(args.gpu_ids)
        else:
            if args.gpu_ids is not None:
                log("Config", "GPU is not available, use CPU for training and ignore '--gpu-ids' setting")

        # If cube id is specified, skip to repeat embedding fit
        if args.cube_id is None or not my_cube.load(embedding_path):
            with ParallelManager():
                if args.input is None:
                    log("Config", "No input file specified")
                    return

                checkExist(args.input)
                mkDirs(embedding_path)

                log("Preprocess", "Load data")
                label = None
                if args.input.endswith(".h5"):
                    samples = pd.read_hdf(args.input)

                elif args.input.endswith(".h5ad"):
                    samples = sc.read_h5ad(args.input)
                    if "type" in samples.obs:
                        label = (
                            samples.obs["type"] == "doublet").astype("int8")

                    samples = pd.DataFrame(samples.X.toarray() if issparse(
                        samples.X) else samples.X,
                                           index=samples.obs_names,
                                           columns=samples.var_names)

                else:
                    raise NotImplementedError("Unsupport file format")

                # if real label not specific, disable validation
                if label is None:
                    label = pd.Series(np.zeros_like(samples.index,
                                                    dtype="int8"),
                                      index=samples.index,
                                      name="type")
                    if args.enable_validation:
                        log(
                            "Config",
                            "Disable '--enable-validation' because no real label found"
                        )
                        args.enable_validation = False

                label.to_csv(join(embedding_path, "ExperimentLabel.csv"),
                             header=None)

                checkData(samples)
                future: Future = createTrainData(samples,
                                                 output_path=embedding_path,
                                                 adj=args.adj_factor,
                                                 seed=args.seed)

                samples = samples.T
                writeHdf(
                    samples,
                    join(
                        embedding_path,
                        f"00-dataByGene[{samples.dtypes.iloc[0].name}][raw].h5"
                    ))

                log("Pre-processing", "Filter data")
                samples = filterData(samples)
                writeHdf(
                    samples,
                    join(
                        embedding_path,
                        "01-dataByGene[{:s}][filter].h5".format(
                            samples.dtypes.iloc[0].name)))

                my_cube.fit(samples,
                            device=gpu_ids[0] if isinstance(gpu_ids, list) and len(gpu_ids)>0 else "cpu",
                            seed=args.seed,
                            latent_dim=args.dim,
                            job_id=cube_id)

            train_data, train_label = future.result()
            log("Post-processing",
                "Processing data with log, std and feature minmax")
            train_data = minmax(std(np.log(train_data + 1)))
            log("Post-processing", "Single channels data is tranforming")

            my_cube.batchTransform(train_data, train_path)
            writeCsv(train_label,
                     join(train_path, "TrainLabel.csv"),
                     header=None)
            if checkExist(join(embedding_path, "ExperimentLabel.csv"),
                          raise_error=False):
                shutil.copyfile(join(embedding_path, "ExperimentLabel.csv"),
                                join(train_path, "ExperimentLabel.csv"))

        elif args.input is not None:
            log("Config", "input is ignored because cube id is specified")

        log("Train", "Data check before training start")
        dim = my_cube._config["latent_dim"]
        checkExist(join(train_path, "TrainLabel.csv"))
        checkShape(train_path, shape=(args.dim, None, None))
        log("Train", f"Data check passed, data path {train_path}")
        model_id = fit(
            home_dir=home_path,
            data_dir=train_path,
            lr=args.learning_rate,
            gamma=0.99,
            epochs=args.epochs,
            train_batch=args.train_batch_size,
            valid_batch=args.vaild_batch_size,
            in_channels=dim,
            transform=None,
            shuffle=True,
            gpu_ids=gpu_ids,
            seed=args.seed,
            label_file="TrainLabel.csv",
            threshold=args.threshold,
            k=args.k,
            once=False,
            use_index=False,
            step=5,
            max_acc_limit=1,
            multi_process=args.enable_multiprocess)

        log("Inference", "Begin doublet detection output")
        infer(data_dir=train_path,
              home_dir=home_path,
              model_id=model_id,
              label_file="ExperimentLabel.csv",
              in_channels=dim,
              k=args.k,
              threshold=args.threshold,
              batch_size=args.infer_batch_size,
              gpu_ids=gpu_ids,
              with_eval=args.enable_validation,
              seed=args.seed,
              multi_process=args.enable_multiprocess)

        em.setNormalInfo("Doublet detection finished")



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch
import numpy as np

from socube.utils.logging import log

__all__ = ["EarlyStopping"]


class EarlyStopping(object):
    """
    Early stops the training if validation loss doesn't improve after a given patience.

    Parameters
    ----------
    descend: bool, default True
        If True, the loss is minimized. If False, the loss is maximized.
    patience: int, default 7
        How long to wait after last time validation loss improved.
    verbose: int, default 0
        Prints a message which level great than `verbose` for each
         validation loss improvement.
    delta: float, default 0
        Minimum change in the monitored quantity to qualify as an improvement.
    path: str, default 'checkpoint.pt'
        Path for the checkpoint to be saved to.
    """

    def __init__(self,
                 descend: bool = True,
                 patience: int = 7,
                 threshold: float = 1e-5,
                 verbose: int = 0,
                 delta: float = 0,
                 path='checkpoint.pt'):
        self._flag = 1 if descend else -1
        self._patience = patience
        self._verbose = verbose
        self._counter = 0
        self._best = np.inf if descend else -np.inf
        self._stop = False
        self._delta = delta
        self._path = path
        self._threshold = threshold

    def __call__(self,
                 score: float,
                 model: torch.nn.Module,
                 flag: bool = True) -> bool:
        """
        Object callable function to update record

        Parameters
        --------------
        score: float
            the loss score of model
        model: torch.nn.Module
            the model to be training
        flag: bool, default True
            If True, it will skipped at this time

        Returns
        --------------
            Boolean value
        """
        if flag and (self._best + self._delta - score) * self._flag > 0:
            log(EarlyStopping.__name__,
                f"Score changes from {self._best} to {score}",
                quiet=self._verbose > 1)
            torch.save(model.state_dict(), self._path)
            self._best = score
            self._counter = 0
            return True
        else:
            self._counter += 1
            log(EarlyStopping.__name__,
                f'EarlyStopping counter: {self._counter} out of {self._patience}',
                quiet=self._verbose > 5)
            if self._counter >= self._patience:
                self._stop = True
            return False

    @property
    def earlyStop(self) -> bool:
        """
        Wether to reach early stopping point.

        Returns
        --------------
            Boolean value
        """
        return self._stop or (self._best - self._threshold) * self._flag <= 0



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from types import TracebackType
from typing import Any, Type
from abc import ABCMeta, abstractmethod

__all__ = ["ContextManagerBase"]


class ContextManagerBase(metaclass=ABCMeta):
    """
    Basical interface for context manager.
    """

    @abstractmethod
    def __enter__(self):
        """
        Enter the context.
        """
        return self

    @abstractmethod
    def __exit__(self, exc_type: Type[Exception], exc_obj: Exception,
                 exc_trace: TracebackType) -> bool:
        """
        __exit__ method's implemation for context manager, python interpreter
        will automatically catch exception information and transfer to this
        method as parameters. If parameters are all equal with 'None', it means
        no exception occurred and you are very lucky.

        Parameters
        ----------
        exc_type : Type[Exception]
            The type of exception.
        exc_obj : Exception
            The exception object.
        exc_trace : TracebackType
            The traceback object of the exception.

        Returns
        -------
        bool, True if the exception is handled, False otherwise.
        """
        return False

    def __call__(self, func):
        """
        As a decorator to enable the context manager.
        """

        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with self:
                return func(*args, **kwargs)

        return wrapper



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from setuptools import setup, find_packages

install_requires = [
    "anndata>=0.7.8",
    "matplotlib>=3.5.0",
    "numpy>=1.20.1",
    "pandas>=1.3.4",
    "python-highcharts>=0.4.2",
    "scanpy>=1.8.2",
    "scikit-learn>=1.0.1",
    "scipy>=1.7.3",
    "tables<=3.6.1",
    "lapjv>=1.3.1",
    "torch>=1.8.1",
    "torchvision>=0.9.1",
    "tqdm>=4.62.3",
    "umap-learn>=0.5.2"
]

setup(
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    platforms="any",
    install_requires=install_requires,
    entry_points=dict(console_scripts=["socube = socube:main"]),
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



# MIT License
#
# Copyright (c) 2022 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch
import torch.nn as nn

from ._base import ModuleBase

__all__ = [
    "InceptionBlock2D", "InceptionBlock2DLayer"
]


class InceptionBlock2D(ModuleBase):
    """
    Inception block with multiple conv kernels

    Parameters
    ----------
    in_channels: integer scalar value
        input data's channel count
    out_channels: integer scalar value
        output data's channel count,
        it must be a multiple of `num`
    nums: integer scalar value
        Number of convolution kernels

    Examples
    ----------
    >>> iblock = InceptionBlock2D(48, 96, 3)
    """
    def __init__(self, in_channels: int, out_channels: int, nums: int = 3):
        super(InceptionBlock2D, self).__init__(in_channels, out_channels)
        assert isinstance(nums, int) and nums > 0
        assert out_channels % nums == 0, f"Out channels should be multiple of {nums}"
        out_channels //= nums
        self._inceptions = nn.ModuleList()
        for padding in range(nums):
            self._inceptions.append(
                nn.Sequential(
                    nn.Conv2d(in_channels,
                              out_channels,
                              2 * padding + 1,
                              padding=padding),
                    nn.BatchNorm2d(out_channels),
                    nn.ReLU(inplace=True)))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.cat([inception(x) for inception in self._inceptions],
                         dim=1)


class InceptionBlock2DLayer(ModuleBase):
    """
    Inception block with multiple layers.

    Parameters
    ----------
    in_channels: integer scalar value
        input data's channel count
    out_channels: integer scalar value
        output data's channel count
    """
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super(InceptionBlock2DLayer, self).__init__(in_channels, out_channels)
        self.__extracts = nn.Sequential(
            nn.Conv2d(in_channels, 48, 3, padding=1),
            nn.BatchNorm2d(48),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(3, 2, padding=1),
            InceptionBlock2D(48, 96),
            nn.MaxPool2d(3, 2, padding=1),
            nn.Dropout(0.1),
            InceptionBlock2D(96, out_channels),
            nn.AdaptiveMaxPool2d((1, 1)),
            nn.Flatten())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.__extracts(x)
